/* demo_xy.c
 * 
 * Copyright (C) 2024 Marcin Kulbaka
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>

#include "lbmn_relax_xy_D.h"
#include "lbmn_mesh_xy_D.h"
#include "lbmn_anls_xy_D.h"

/* problems */
#include "demo_xy_problems_1.h"
#include "demo_xy_problems_2.h"

/*
* (maximal) number of electrodes (colours on workbench picture)
*
* [0] - vacuum
* [1] - electrode No. 1
* [2] - electroce No. 2 etc.
* ...
* [NCOLOURS - 1] - last electrode
*/
#define N_ELECTRODE_COLOURS  6

#define ITER_MAX 1000000
#define EPS (1.0e-9)


/*
* (maximal number of isolines (V)
* [0] - empty
* [1] - isoline 1
* [2] - isoline 2
* [...]
* [N_ISO_V - 1] - last isoline
*
* isolines (V)  1 .. 5
*
*/
#define N_ISOLINES_V 6


/* pointer to problem defining function */
typedef int (*T_fptr_lbmn_problem_xy)(
			unsigned int* ptr_size_row,/*output*/
            unsigned int* ptr_size_col,
            double* ptr_x_start,
            double* ptr_x_stop,
            double* ptr_y_start,
            double* ptr_y_stop,
			
			double** ptrptr_V,
			unsigned int** ptrptr_Bnd,
			double** ptrptr_E_x,
			double** ptrptr_E_y,
			double** ptrptr_E_1,
			double** ptrptr_E_2,
            unsigned int** ptrptr_isolines_V,
            unsigned int** ptrptr_isolines_E_x,
            unsigned int** ptrptr_isolines_E_y,
            unsigned int** ptrptr_isolines_E_1,
            unsigned int** ptrptr_isolines_E_2,
			unsigned char** ptrptr_workbench_colours_r,
			unsigned char** ptrptr_workbench_colours_g,
			unsigned char** ptrptr_workbench_colours_b,
			unsigned char** ptrptr_workbench_bw_colours_r,
			unsigned char** ptrptr_workbench_bw_colours_g,
			unsigned char** ptrptr_workbench_bw_colours_b);



/*=== colours of electrodes ===*/

/* colours of electrodes for workbench (colours) */
unsigned char tab_workbench_electrode_colours_red   [N_ELECTRODE_COLOURS];
unsigned char tab_workbench_electrode_colours_green [N_ELECTRODE_COLOURS];
unsigned char tab_workbench_electrode_colours_blue  [N_ELECTRODE_COLOURS];


/*=== colours of electrodes for other pictures (V, E_x, E_y, E_1, E_2)) ===*/
unsigned char tab_electrode_colours_red   [N_ELECTRODE_COLOURS];
unsigned char tab_electrode_colours_green [N_ELECTRODE_COLOURS];
unsigned char tab_electrode_colours_blue  [N_ELECTRODE_COLOURS];


/*=== isolines V_xy - values ===*/
double tab_isoline_V_values [N_ISOLINES_V];


/*=== colours of isolines (V) ===*/

/* version for black-white workbench */
unsigned char tab_isoline_bw_V_colours_red   [N_ISOLINES_V];
unsigned char tab_isoline_bw_V_colours_green [N_ISOLINES_V];
unsigned char tab_isoline_bw_V_colours_blue  [N_ISOLINES_V];


/* version for colorful workbench (hot-to-cold) */
unsigned char tab_isoline_c_V_colours_red   [N_ISOLINES_V];
unsigned char tab_isoline_c_V_colours_green [N_ISOLINES_V];
unsigned char tab_isoline_c_V_colours_blue  [N_ISOLINES_V];


/*=== filenames ===*/

const char* filename_V_xy_dat           		= "./outputs_xy/V_xy.dat";
const char* filename_V_xy_gray          		= "./outputs_xy/V_xy_gray.ppm";
const char* filename_V_xy_gray_inv      		= "./outputs_xy/V_xy_gray_inv.ppm";
const char* filename_V_xy_hottocold_clamp  		= "./outputs_xy/V_xy_hottocold.ppm";
const char* filename_V_xy_hottocold_clamp_inv 	= "./outputs_xy/V_xy_hottocold_inv.ppm";
const char* filename_V_xy_jet_clamp     		= "./outputs_xy/V_xy_jet.ppm";
const char* filename_V_xy_jet_clamp_inv 		= "./outputs_xy/V_xy_jet_inv.ppm";

const char* filename_E_x_dat					= "./outputs_xy/E_x.dat";
const char* filename_E_x_gray          			= "./outputs_xy/E_x_gray.ppm";
const char* filename_E_x_gray_inv      			= "./outputs_xy/E_x_gray_inv.ppm";
const char* filename_E_x_hottocold_clamp     	= "./outputs_xy/E_x_hottocold.ppm";
const char* filename_E_x_hottocold_clamp_inv 	= "./outputs_xy/E_x_hottocold_inv.ppm";
const char* filename_E_x_jet_clamp     			= "./outputs_xy/E_x_jet.ppm";
const char* filename_E_x_jet_clamp_inv 			= "./outputs_xy/E_x_jet_inv.ppm";

const char* filename_E_y_dat					= "./outputs_xy/E_y.dat";
const char* filename_E_y_gray          			= "./outputs_xy/E_y_gray.ppm";
const char* filename_E_y_gray_inv      			= "./outputs_xy/E_y_gray_inv.ppm";
const char* filename_E_y_hottocold_clamp     	= "./outputs_xy/E_y_hottocold.ppm";
const char* filename_E_y_hottocold_clamp_inv	= "./outputs_xy/E_y_hottocold_inv.ppm";
const char* filename_E_y_jet_clamp     			= "./outputs_xy/E_y_jet.ppm";
const char* filename_E_y_jet_clamp_inv 			= "./outputs_xy/E_y_jet_inv.ppm";

const char* filename_E_1_dat					= "./outputs_xy/E_1.dat";
const char* filename_E_1_gray          			= "./outputs_xy/E_1_gray.ppm";
const char* filename_E_1_gray_inv      			= "./outputs_xy/E_1_gray_inv.ppm";
const char* filename_E_1_hottocold_clamp     	= "./outputs_xy/E_1_hottocold.ppm";
const char* filename_E_1_hottocold_clamp_inv 	= "./outputs_xy/E_1_hottocold_inv.ppm";
const char* filename_E_1_jet_clamp     			= "./outputs_xy/E_1_jet.ppm";
const char* filename_E_1_jet_clamp_inv 			= "./outputs_xy/E_1_jet_inv.ppm";

const char* filename_E_2_dat          			= "./outputs_xy/E_2.dat";
const char* filename_E_2_gray          			= "./outputs_xy/E_2_gray.ppm";
const char* filename_E_2_gray_inv      			= "./outputs_xy/E_2_gray_inv.ppm";
const char* filename_E_2_hottocold_clamp     	= "./outputs_xy/E_2_hottocold.ppm";
const char* filename_E_2_hottocold_clamp_inv 	= "./outputs_xy/E_2_hottocold_inv.ppm";
const char* filename_E_2_jet_clamp     			= "./outputs_xy/E_2_jet.ppm";
const char* filename_E_2_jet_clamp_inv 			= "./outputs_xy/E_2_jet_inv.ppm";

const char* filename_Bnd_xy_dat         		= "./outputs_xy/Bnd_xy.dat";
const char* filename_workbench_bw         		= "./outputs_xy/workbench_bw.ppm";
const char* filename_workbench_colours    		= "./outputs_xy/workbench_colours.ppm";

const char* filename_anls_V_xy_isolines_bw		= "./outputs_xy/anls_V_xy_isolines_bw.ppm";
const char* filename_anls_V_xy_isolines_colours	= "./outputs_xy/anls_V_xy_isolines_colours.ppm";


/****************** main program *********************/

int
main(int argc, char*argv[])
{
	/* general status of program and return value */
	int error_code = 0;
	
	/* if problem has been dynamically allocated */
	int problem_allocated = 0;
	
	/*==========*
	* MESH DATA *
	*===========*/


	/* mesh size */
	unsigned int size_row    = 0;
	unsigned int size_col    = 0;

	/* mesh data */
	double* ptr_V           = NULL; /* mesh V(x,y) */
    unsigned int* ptr_Bnd   = NULL; /* boundaries (mesh) */
	
	double* ptr_E_x 		= NULL; /* mesh E_x */
	double* ptr_E_y 		= NULL; /* mesh E_y */

	double* ptr_E_1 		= NULL; /* mesh E^1*/
	double* ptr_E_2	 		= NULL; /* mesh E^2 */

    unsigned int* ptr_isolines_V    = NULL;
    unsigned int* ptr_isolines_E_x  = NULL;
    unsigned int* ptr_isolines_E_y  = NULL;
    unsigned int* ptr_isolines_E_1  = NULL;
    unsigned int* ptr_isolines_E_2  = NULL;
	
	/* workbench images (meshes of pixels) */
	unsigned char* ptr_workbench_colours_red   = NULL;
	unsigned char* ptr_workbench_colours_green = NULL;
	unsigned char* ptr_workbench_colours_blue  = NULL;
	
	unsigned char* ptr_workbench_bw_red   = NULL;
	unsigned char* ptr_workbench_bw_green = NULL;
	unsigned char* ptr_workbench_bw_blue  = NULL;


	/* pointers to electrode colours (workbench - colours)*/
	
    unsigned char* ptr_workbench_electrode_colours_red   = NULL;
	unsigned char* ptr_workbench_electrode_colours_green = NULL;
	unsigned char* ptr_workbench_electrode_colours_blue  = NULL;
    

	/* pointers to electrode colours on V, E_x, E_y, E_1 and E_2 pictures */
	unsigned char* ptr_electrode_colours_red   = NULL;
	unsigned char* ptr_electrode_colours_green = NULL;
	unsigned char* ptr_electrode_colours_blue  = NULL;



	/*=== pointers to isoline colours  ===*/
	
	/* black - white version */
	unsigned char* ptr_isoline_colours_bw_red   = NULL;
	unsigned char* ptr_isoline_colours_bw_green = NULL;
	unsigned char* ptr_isoline_colours_bw_blue  = NULL;
	
	/* colour version */
/*
	unsigned char* ptr_isoline_colours_c_red   = NULL;
	unsigned char* ptr_isoline_colours_c_green = NULL;
	unsigned char* ptr_isoline_colours_c_blue  = NULL;
*/

	/*=== colours for "in range colormaps" ===*/

	/* colour - below range */	
	unsigned char colour_below_red 		= 0; /* black */
	unsigned char colour_below_green 	= 0;
	unsigned char colour_below_blue 	= 0;
	
	/* colour - abowe range */
	unsigned char colour_abowe_red 		= 255;/* white */
	unsigned char colour_abowe_green 	= 255;
	unsigned char colour_abowe_blue 	= 255;


	/*=== colour of background for isoline ===*/
	/* bw */
	unsigned char colour_isolines_background_bw_red 	= 25;/* dark blue */
	unsigned char colour_isolines_background_bw_green 	= 35;
	unsigned char colour_isolines_background_bw_blue 	= 230;

	
	
	unsigned int number_of_electrodes = 0;
	
	unsigned int number_of_isolines_V = N_ISOLINES_V - 1;
	
	
	double x_min = 0.0;
	double x_max = 0.0;
	double y_min = 0.0;
	double y_max = 0.0;

	double mesh_step = 1.0;


	/* min / max values of computed data */
    double V_min = 0.0;
    double V_max = 0.0;
	
	double E_x_min = 0.0;
	double E_x_max = 0.0;
	
	double E_y_min = 0.0;
	double E_y_max = 0.0;
	
	double E_1_min = 0.0;
	double E_1_max = 0.0;
	
	double E_2_min = 0.0;
	double E_2_max = 0.0;

	/* iteration counters */
    unsigned int i_iter = 0;
    unsigned int i_iter_max = ITER_MAX;
    
	unsigned int counter_1000 = 0;/* to display each 1000-th eps on display */

	/* temporary error of solver*/
    double eps_temp = 0.0;
    int ret_val = 0;
	
    int ret_val_workbench_bw        = 0;
    int ret_val_workbench_colours   = 0;

    
    /* pointer to problem definig function */
    T_fptr_lbmn_problem_xy fptr_problem;

    int i_problem = 0; /* index of problem, 1,2, .. */

	/*=========================================* 
	* colours of nodes for workbench (colours) *
	*==========================================*/

	/* set maximal number of electrodes */
	number_of_electrodes = N_ELECTRODE_COLOURS;

	
	/* 0 - vacuum -> white	*/
	tab_workbench_electrode_colours_red		[0] = 255;
	tab_workbench_electrode_colours_green	[0] = 255;
	tab_workbench_electrode_colours_blue  	[0] = 255;
	
	/* 1 -> red 			*/
	tab_workbench_electrode_colours_red		[1] = 128;
	tab_workbench_electrode_colours_green	[1] = 0;
	tab_workbench_electrode_colours_blue	[1] = 0;
	
	/* 2 -> green			*/
	tab_workbench_electrode_colours_red		[2] = 0;
	tab_workbench_electrode_colours_green	[2] = 128;
	tab_workbench_electrode_colours_blue	[2] = 0;

	/* 3 -> blue			*/
	tab_workbench_electrode_colours_red		[3] = 0;
	tab_workbench_electrode_colours_green	[3] = 0;
	tab_workbench_electrode_colours_blue	[3] = 128;
	
    /* 4 -> yellow			*/
	tab_workbench_electrode_colours_red		[4] = 255;
	tab_workbench_electrode_colours_green	[4] = 255;
	tab_workbench_electrode_colours_blue	[4] = 0;
	
    /* 5 -> orange			*/
	tab_workbench_electrode_colours_red		[5] = 245;
	tab_workbench_electrode_colours_green	[5] = 128;
	tab_workbench_electrode_colours_blue	[5] = 0;
	
	
	/*==========================================* 
	* colours of electrodes for others pictures *
	*===========================================*/
	
	/* 0 -> vacuum (not used) */
	tab_electrode_colours_red   [0] = 0;
	tab_electrode_colours_green [0] = 0;
	tab_electrode_colours_blue  [0] = 0;
	
	/* 1 -> black */
	tab_electrode_colours_red   [0] = 0;
	tab_electrode_colours_green [0] = 0;
	tab_electrode_colours_blue  [0] = 0;
	
	/* 2 -> black */
	tab_electrode_colours_red   [0] = 0;
	tab_electrode_colours_green [0] = 0;
	tab_electrode_colours_blue  [0] = 0;
	
	/* 3 -> black */
	tab_electrode_colours_red   [0] = 0;
	tab_electrode_colours_green [0] = 0;
	tab_electrode_colours_blue  [0] = 0;	
	
	
	/*========================================*
	* colours of isolines (V) - black/white   *
	*=========================================*/
	
	/* 0 -> white */
	tab_isoline_bw_V_colours_red   [0] = 255;
	tab_isoline_bw_V_colours_green [0] = 255;
	tab_isoline_bw_V_colours_blue  [0] = 255;

	/* 1 -> yellow */
	tab_isoline_bw_V_colours_red   [1] = 0;
	tab_isoline_bw_V_colours_green [1] = 250;
	tab_isoline_bw_V_colours_blue  [1] = 0;

	/* 2 -> yellow */
	tab_isoline_bw_V_colours_red   [2] = 0;
	tab_isoline_bw_V_colours_green [2] = 250;
	tab_isoline_bw_V_colours_blue  [2] = 0;

	/* 3 -> yellow */
	tab_isoline_bw_V_colours_red   [3] = 0;
	tab_isoline_bw_V_colours_green [3] = 250;
	tab_isoline_bw_V_colours_blue  [3] = 0;

	/* 4 -> yellow */
	tab_isoline_bw_V_colours_red   [4] = 0;
	tab_isoline_bw_V_colours_green [4] = 250;
	tab_isoline_bw_V_colours_blue  [4] = 0;

	/* 5 -> yellow */
	tab_isoline_bw_V_colours_red   [5] = 0;
	tab_isoline_bw_V_colours_green [5] = 250;
	tab_isoline_bw_V_colours_blue  [5] = 0;



	/*========================================*
	* colours of isolines (V) - colour        *
	*=========================================*/
	
	/* 0 -> white */
	tab_isoline_c_V_colours_red   [0] = 255;
	tab_isoline_c_V_colours_green [0] = 255;
	tab_isoline_c_V_colours_blue  [0] = 255;

	/* 1 -> magenta */
	tab_isoline_c_V_colours_red   [1] = 250;
	tab_isoline_c_V_colours_green [1] = 0;
	tab_isoline_c_V_colours_blue  [1] = 250;

	/* 2 -> magenta */
	tab_isoline_c_V_colours_red   [2] = 250;
	tab_isoline_c_V_colours_green [2] = 0;
	tab_isoline_c_V_colours_blue  [2] = 250;

	/* 3 -> magenta */
	tab_isoline_c_V_colours_red   [3] = 250;
	tab_isoline_c_V_colours_green [3] = 0;
	tab_isoline_c_V_colours_blue  [3] = 250;

	/* 4 -> magenta */
	tab_isoline_c_V_colours_red   [4] = 250;
	tab_isoline_c_V_colours_green [4] = 0;
	tab_isoline_c_V_colours_blue  [4] = 250;

	/* 5 -> magenta */
	tab_isoline_c_V_colours_red   [5] = 250;
	tab_isoline_c_V_colours_green [5] = 0;
	tab_isoline_c_V_colours_blue  [5] = 250;



	/*==============================*
	* link pointers to data tables  * 
	*===============================*/



	/*=== set pointers to electrode colours ===*/

    /* colors of electrodes for workbench (colours) */
    ptr_workbench_electrode_colours_red   = tab_workbench_electrode_colours_red;
    ptr_workbench_electrode_colours_green = tab_workbench_electrode_colours_green;
    ptr_workbench_electrode_colours_blue  = tab_workbench_electrode_colours_blue;

    
    /* colours of electrodes for other pictures */
	ptr_electrode_colours_red   = tab_electrode_colours_red;
	ptr_electrode_colours_green = tab_electrode_colours_green;
	ptr_electrode_colours_blue  = tab_electrode_colours_blue; 



	/*=== set pointers to isoline V colours ===*/

	/* black - white version */
	ptr_isoline_colours_bw_red   = tab_isoline_bw_V_colours_red;
	ptr_isoline_colours_bw_green = tab_isoline_bw_V_colours_green;
	ptr_isoline_colours_bw_blue  = tab_isoline_bw_V_colours_blue;

	/* colour version */
/*	
	ptr_isoline_colours_c_red   = tab_isoline_c_V_colours_red;
	ptr_isoline_colours_c_green = tab_isoline_c_V_colours_green;
	ptr_isoline_colours_c_blue  = tab_isoline_c_V_colours_blue;
*/

	/*
	* print hello message to user
	*/
	
    printf("+------------------------------+\n");
    printf("|        Liebmann DEMO_XY      |\n");
    printf("+------------------------------+\n");
    printf("Marcin Kulbaka 2024\n");
    printf("license: GNU GPL v3.0+\n\n");
	
	
    if(argc != 2)
    {/* argc != 2 */

        printf("ERROR - incorrect number of program parameters\n");
        printf("using: DEMO_XY i, where i = 1..n (index of problem to solve)\n");

		/* failure No 1 */
        error_code = -1;
		goto LABEL_ABEND;
		
    }/* argc != 2 */

    /*
	* else argv == 2 
	*/
  
  
    /*
	*	set problem to load 
	*/
    i_problem = atoi(argv[1]);
   
    switch(i_problem)
    {
        case 1:
        fptr_problem = lbmn_demo_xy_problem_1_2_plates_200;
        break;

        case 2:
        fptr_problem = lbmn_demo_xy_problem_2_2_plates_400;
        break;

        case 3:
        fptr_problem = lbmn_demo_xy_problem_3_flat_diode_201;
        break;

        case 4:
        fptr_problem = lbmn_demo_xy_problem_4_S_rectangle_500_300;
        break;
		
		case 5:
		fptr_problem = lbmn_demo_xy_problem_5_electrolytic_tank_601_401;
		break;

		case 6:
		fptr_problem = lbmn_demo_xy_problem_6_dualsegment_magnetron_smile_401_401;
		break;

        case 7:
        fptr_problem = lbmn_demo_xy_problem_7_dualsegment_magnetron_Paszkowski_p62_401_401;
        break;

        case 8:    
        fptr_problem = lbmn_demo_xy_problem_8_flat_diaphragm_Paszkowski_p62_401_401;
        break;

        case 9:
        fptr_problem = lbmn_demo_xy_problem_9_2_inequal_plates_200;
        break;

        case 10:
        fptr_problem = lbmn_demo_xy_problem_10_modified_flat_diaphragm_Paszkowski_401_401;
        break;

		case 11:
        fptr_problem = lbmn_demo_xy_problem_11_electrolytic_tank_inequal_601_401;
        break;

		case 12:
        fptr_problem = lbmn_demo_xy_problem_12_electrolytic_tank_equal_pots_601_401;
        break;
		
		case 13:
        fptr_problem = lbmn_demo_xy_problem_13_2_inequal_plates_double_400;
        break;
		
		case 14:
		fptr_problem = lbmn_demo_xy_problem_14_quadrupole_401_401;
		break;

		case 15:
		fptr_problem = lbmn_demo_xy_problem_15_cylindrical_capacitor_401_401;
		break;

		case 16:
		fptr_problem = lbmn_demo_xy_problem_16_quadrupole_v2_401_401;
		break;

		case 17:
		fptr_problem = lbmn_demo_xy_problem_17_hexapole_401_401;
		break;

		case 18:
		fptr_problem = lbmn_demo_xy_problem_18_skew_quadrupole_401_401;
		break;

		case 19:
		fptr_problem = lbmn_demo_xy_problem_19_octupole_401_401;
		break;

		case 20:
		fptr_problem = lbmn_demo_xy_problem_20_quadrupole_plates_401_401;
		break;
				
        /* othertwise - error */
        default:
        printf("incorrect index of problem\n");
        printf("program terminated\n");
        error_code = -2;
		goto LABEL_ABEND;
		
    }/* switch problem */

    

    /*
    * try to initialize problem 
    */

    ret_val = (*fptr_problem)(
			    & size_row,/*output*/
                & size_col,
                & x_min,
                & x_max,
                & y_min,
                & y_max,
				
			    & ptr_V,
			    & ptr_Bnd,
			    & ptr_E_x,
			    & ptr_E_y,
			    & ptr_E_1,
			    & ptr_E_2,
                & ptr_isolines_V,
                & ptr_isolines_E_x,
                & ptr_isolines_E_y,
                & ptr_isolines_E_1,
                & ptr_isolines_E_2,
			    & ptr_workbench_colours_red,
			    & ptr_workbench_colours_green,
			    & ptr_workbench_colours_blue,
			    & ptr_workbench_bw_red,
			    & ptr_workbench_bw_green,
			    & ptr_workbench_bw_blue);

    printf("memory allocation ret_val == %i\n", ret_val);
    printf("ptr_V   == %p\n", (void*) ptr_V);
    printf("ptr_Bnd == %p\n", (void*) ptr_Bnd);

    if(ret_val != 0)
    {/* problem with defining problem to solve */

        printf("ERROR - could not create problem to solve.\n");
        printf("error code == %i\n", ret_val);

        error_code = -3;
		goto LABEL_ABEND;
		
    }/* problem with defining problem to solve*/
    else
    {/* problem to solve created */


        printf("problem to solve has been created \n");
		
		/* mark for further using after LABEL_x */
		problem_allocated = 1;
		

	   /*
		* print problem
		*/
		
		printf("size_row == %u\n", size_row);
		printf("size_col == %u\n", size_col);

		
		printf("\n");
		printf("V on mesh:\n\n");
		lbmn_mesh_xy_D_printf_V(ptr_V,
								 size_row,
								 size_col);

		printf("\n");
		printf("boundaries on mesh:\n\n");
		lbmn_mesh_xy_D_printf_Bnd(ptr_Bnd,
								   size_row,
								   size_col);




		/* === SOLVE LAPLACE === */

		printf("\n>>> Solving \n\n");

		i_iter = 1;
		counter_1000 = 1;


		eps_temp = lbmn_relax_xy_D_iter(ptr_V, ptr_Bnd, size_row, size_col);

		printf("after first iter\n");

		printf("%7d ---  %+18.15E\n", i_iter, eps_temp);

		while(
				(eps_temp > EPS)
				&&
				(i_iter < i_iter_max)
			 )
		{
			eps_temp = lbmn_relax_xy_D_iter(ptr_V, ptr_Bnd, size_row, size_col);

			i_iter++;

			counter_1000 ++;
			
			if(counter_1000 >= 1000)
			{
				counter_1000 = 0; /*reset counter*/
				
				/*display another 1000-th eps */
				printf("%7d ---  %+18.15E\n", i_iter, eps_temp);
			}

		}/* loop iter relax */
		
		
		
		
		if(i_iter >= i_iter_max)
		{/* Could not find Laplace solution */
	
			printf("\n\n");
			printf("*** FAILURE ***\n");
			printf("Could not find Laplace solution\n");
			
			error_code = -4;
			goto LABEL_ABEND;
			
		}/* Couldn not find Laplacec solution */
		else
		{/* Laplace solution has been found */    

			/* === end of relaxation, solution found (+/-)  === */

			printf("\n\n");

			printf("*** SUCCESS ***\n");
			printf("solution V(x,y) has been found\n");
			printf("after %d iterations\n\n", i_iter);

			printf("V calculated\n");

			lbmn_mesh_xy_D_find_V_min_max(
								&V_min,
								&V_max,
								ptr_V,  
								size_row,
								size_col);


			printf("V_min == %+f\n", V_min);
			printf("V_max == %+f\n", V_max);


			/*
			* try to write V_xy to dat and PPM files (with appropriate colormaps)
			*/
					
			if(		lbmn_mesh_xy_D_fprintf_mesh_double_dat(
					filename_V_xy_dat,
					ptr_V, 
					size_row,/* size of mesh */
					size_col)
				||
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_gray,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_grayscale,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)
				||
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_gray_inv,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_grayscale_inv,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)
				|| 
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_hottocold_clamp,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_hottocold_clamp,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)

				||
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_hottocold_clamp_inv,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_hottocold_clamp_inv,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)
				
				||
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_jet_clamp,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_jet_clamp,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)
				
				||
					lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
					filename_V_xy_jet_clamp_inv,
					ptr_V,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col,
					lbmn_colormap_jet_clamp_inv,
					V_min,/* range of displayed value */
					V_max,
					colour_below_red,/* colour - below range */
					colour_below_green,
					colour_below_blue,
					colour_abowe_red,/* colour - abowe range */
					colour_abowe_green,
					colour_abowe_blue,
					ptr_electrode_colours_red,
					ptr_electrode_colours_green,
					ptr_electrode_colours_blue,
					number_of_electrodes)
				)
			{
				printf("ERROR - could not write V_xy data to files:\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
							filename_V_xy_dat, filename_Bnd_xy_dat,
							filename_V_xy_gray, filename_V_xy_gray_inv,
							filename_V_xy_hottocold_clamp, filename_V_xy_hottocold_clamp_inv,
							filename_V_xy_jet_clamp, filename_V_xy_jet_clamp_inv);
							
				error_code = -5;
				goto LABEL_ABEND;
			}
			else
			{
				printf("All V_xy files written successfully\n");
			}
			
			
			
			/*
			*=== try to create workbench images (bw/colours) ===
			*/
			
			ret_val_workbench_bw = 	lbmn_mesh_xy_D_create_image_workbench_bw(
					ptr_workbench_bw_red,/* targets */
					ptr_workbench_bw_green,
					ptr_workbench_bw_blue,
					size_row,/* size of mesh */
					size_col,
					ptr_Bnd);/* mesh of boundaries */

			ret_val_workbench_colours=	lbmn_mesh_xy_D_create_image_workbench_colours(
					ptr_workbench_colours_red,/* targets */
					ptr_workbench_colours_green,
					ptr_workbench_colours_blue,
					size_row,/* size of mesh */
					size_col,
					ptr_Bnd,/* mesh of boundaries*/
					ptr_workbench_electrode_colours_red,/* colours of each electrode (Bnd = 1,2,3..) (r/g/b) */
					ptr_workbench_electrode_colours_green,
					ptr_workbench_electrode_colours_blue,
					number_of_electrodes);

			if(ret_val_workbench_bw ||
			   ret_val_workbench_colours
			)

			{/* failed to create workbench images */
			
				printf("ERROR!!! Could not create workbench images!!!\n");
				printf("ret_val_workbench_bw      == %d\n", ret_val_workbench_bw);
				printf("ret_val_workbench_colours == %d\n", ret_val_workbench_colours);

				error_code = -6;
				goto LABEL_ABEND;
				
			}/* failed to create workbench images */
			else
			{/* created workbench images */
		
				/* try to write workbench images to files */
				
				if(
					lbmn_mesh_xy_D_fprintf_mesh_uint_dat(
					filename_Bnd_xy_dat,
					ptr_Bnd,
					size_row,/* size of matrices */
					size_col)
				||	
					lbmn_mesh_xy_D_fprintf_image_workbench_netpbm_P3(
					filename_workbench_colours,
					ptr_workbench_colours_red,/* 3 input r/g/b channels */
					ptr_workbench_colours_green,
					ptr_workbench_colours_blue,  
					size_row,/* size of these r/g/b channels */
					size_col)
				||	
					lbmn_mesh_xy_D_fprintf_image_workbench_netpbm_P3(
					filename_workbench_bw,
					ptr_workbench_bw_red,/* 3 input r/g/b channels */
					ptr_workbench_bw_green,
					ptr_workbench_bw_blue,  
					size_row,/* size of these r/g/b channels */
					size_col)						
				)
				{/* error */
					
					printf("ERROR!!! Could not write workbench images to files!!!\n");
					error_code = -7;
					goto LABEL_ABEND;
					
				}/* error */
				else
				{/* success */
				
					printf("All workbench written successfully\n");
				
				}/* success */
				
			}/* created workbench images */


		
			
			
			/*==== Process E_x ====*/
		
			/* 
			* determine x-component of electric force E
			* E_x =  d1V/dx1
			*/
			ret_val = lbmn_mesh_xy_D_calculate_tab_d1V_dx1(
								ptr_E_x,
								ptr_V,
								size_row,
								size_col,
								mesh_step);
								
			if(ret_val != 0)
			{
			
				
				printf("ERROR! Could not calculate E_x (d1V_dx1)\nerror code == %d", ret_val);

				error_code = -8;
				goto LABEL_ABEND;
				
			}
			else
			{/* E_x calculated */
				
					/*
					* try to write E_x to dat and PPM files (with appropriate colormaps)
					*/
					
					lbmn_mesh_xy_D_find_V_min_max(
										&E_x_min,
										&E_x_max,
										ptr_E_x,  
										size_row,/* size of mesh */
										size_col);
					
					printf("E_x calculated\n");
					printf("E_x_min == %+f\n", E_x_min);
					printf("E_x_max == %+f\n", E_x_max);
					
					if(
						lbmn_mesh_xy_D_fprintf_mesh_double_dat(
						filename_E_x_dat,
						ptr_E_x, 
						size_row,/* size of mesh */
						size_col)
					||								
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_gray,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_gray_inv,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale_inv,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					|| 
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_hottocold_clamp,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)

					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_hottocold_clamp_inv,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp_inv,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_jet_clamp,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_x_jet_clamp_inv,
						ptr_E_x,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp_inv,
						E_x_min,/* range of displayed value */
						E_x_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					)
				{
					printf("ERROR - could not write E_x data to files:\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
								filename_E_x_dat,
								filename_E_x_gray, filename_E_x_gray_inv,
								filename_E_x_hottocold_clamp, filename_E_x_hottocold_clamp_inv,
								filename_E_x_jet_clamp, filename_E_x_jet_clamp_inv);
								
					error_code = -9;
					goto LABEL_ABEND;
				}
				else
				{
					printf("All E_x files written successfully\n");
				
				}
			}/* E_x calculated */
		

		
			
			/*==== Process E_y ====*/
		
		
			/* 
			* determine y - component of electric force E
			* E_y =  d1V/dy1
			*/
			
			ret_val = lbmn_mesh_xy_D_calculate_tab_d1V_dy1(
								ptr_E_y,
								ptr_V,
								size_row,
								size_col,
								mesh_step);
								
			if(ret_val != 0)
			{
			
				printf("ERROR! Could not calculate E_y (d1V_dy1)\nerror code == %d", ret_val);
				error_code = -10;
				goto LABEL_ABEND;
			
			}
			else
			{/* E_y calculated */
				
					/*
					* try to write E_y to dat and PPM files (with appropriate colormaps)
					*/
		
					lbmn_mesh_xy_D_find_V_min_max(
										&E_y_min,
										&E_y_max,
										ptr_E_y,  
										size_row,/* size of mesh */
										size_col);
					
					printf("E_y calculated\n");
					printf("E_y_min == %+f\n", E_y_min);
					printf("E_y_max == %+f\n", E_y_max);
					
					if(
						lbmn_mesh_xy_D_fprintf_mesh_double_dat(
						filename_E_y_dat,
						ptr_E_y, 
						size_row,/* size of mesh */
						size_col)
					||				
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_gray,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_gray_inv,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale_inv,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					|| 
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_hottocold_clamp,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)

					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_hottocold_clamp_inv,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp_inv,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_jet_clamp,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_y_jet_clamp_inv,
						ptr_E_y,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp_inv,
						E_y_min,/* range of displayed value */
						E_y_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					)
				{
					printf("ERROR - could not write E_y data to files:\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
								filename_E_y_dat,
								filename_E_y_gray, filename_E_y_gray_inv,
								filename_E_y_hottocold_clamp, filename_E_y_hottocold_clamp_inv,
								filename_E_y_jet_clamp, filename_E_y_jet_clamp_inv);
								
					error_code = -11;
					goto LABEL_ABEND;
				}
				else
				{
					printf("All E_y files written successfully\n");
				
				}
				
			}/* E_y calculated */
			
			
			/* 
			* determine electric force E
			* E_1 = sqrt(Ex*Ex + Ey*Ey)
			*/
			
			ret_val = lbmn_mesh_xy_D_calculate_tab_E1(
								ptr_E_1,
								ptr_E_x,
								ptr_E_y,
								size_row,
								size_col);
								
			if(ret_val != 0)
			{
			
				printf("ERROR! Could not calculate E_1\nerror code == %d", ret_val);
				error_code = -12;
				goto LABEL_ABEND;
			
			}
			else
			{/* E_1 calculated */
				
					/*
					* try to write E_1 to dat and PPM files (with appropriate colormaps)
					*/
		
					lbmn_mesh_xy_D_find_V_min_max(
										&E_1_min,
										&E_1_max,
										ptr_E_1,  
										size_row,/* size of mesh */
										size_col);
					
					printf("E_1 calculated\n");
					printf("E_1_min == %+f\n", E_1_min);
					printf("E_1_max == %+f\n", E_1_max);
					
					if(
						lbmn_mesh_xy_D_fprintf_mesh_double_dat(
						filename_E_1_dat,
						ptr_E_1, 
						size_row,/* size of mesh */
						size_col)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_gray,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_gray_inv,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale_inv,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					|| 
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_hottocold_clamp,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)

					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_hottocold_clamp_inv,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp_inv,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_jet_clamp,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_1_jet_clamp_inv,
						ptr_E_1,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp_inv,
						E_1_min,/* range of displayed value */
						E_1_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					)
				{
					printf("ERROR - could not write E_1 data to files:\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
								filename_E_1_dat,
								filename_E_1_gray, filename_E_1_gray_inv,
								filename_E_1_hottocold_clamp, filename_E_1_hottocold_clamp_inv,
								filename_E_1_jet_clamp, filename_E_1_jet_clamp_inv);
								
					error_code = -13;
					goto LABEL_ABEND;
				}
				else
				{
					printf("All E_1 files written successfully\n");
				
				}
			}/* E_1 calculated */
			
			
			/* 
			* determine square of electric force E^2
			* E_2 = (Ex*Ex + Ey*Ey)
			*/
			
			ret_val = lbmn_mesh_xy_D_calculate_tab_E2(
								ptr_E_2,
								ptr_E_x,
								ptr_E_y,
								size_row,
								size_col);
								
			if(ret_val != 0)
			{
			
				printf("ERROR! Could not calculate E_2\nerror code == %d", ret_val);
				error_code = -14;
				goto LABEL_ABEND;
			
			}
			else
			{/* E_2 calculated */
				
					/*
					* try to write E2 to dat and PPM files (with appropriate colormaps)
					*/
					
					lbmn_mesh_xy_D_find_V_min_max(
										&E_2_min,
										&E_2_max,
										ptr_E_2,  
										size_row,/* size of mesh */
										size_col);
					
					printf("E_2 calculated\n");
					printf("E_2_min == %+f\n", E_2_min);
					printf("E_2_max == %+f\n", E_2_max);
					
					if(		
						lbmn_mesh_xy_D_fprintf_mesh_double_dat(
						filename_E_2_dat,
						ptr_E_2, 
						size_row,/* size of mesh */
						size_col)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_gray,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_gray_inv,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_grayscale_inv,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					|| 
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_hottocold_clamp,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)

					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_hottocold_clamp_inv,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_hottocold_clamp_inv,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_jet_clamp,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					
					||
						lbmn_mesh_xy_D_fprintf_mesh_double_netpbm_P3(
						filename_E_2_jet_clamp_inv,
						ptr_E_2,
						ptr_Bnd,
						size_row,/* size of matrices */
						size_col,
						lbmn_colormap_jet_clamp_inv,
						E_2_min,/* range of displayed value */
						E_2_max,
						colour_below_red,/* colour - below range */
						colour_below_green,
						colour_below_blue,
						colour_abowe_red,/* colour - abowe range */
						colour_abowe_green,
						colour_abowe_blue,
						ptr_electrode_colours_red,
						ptr_electrode_colours_green,
						ptr_electrode_colours_blue,
						number_of_electrodes)
					)
				{
					printf("ERROR - could not write E_2 data to files:\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
								filename_E_2_dat,
								filename_E_2_gray, filename_E_2_gray_inv,
								filename_E_2_hottocold_clamp, filename_E_2_hottocold_clamp_inv,
								filename_E_2_jet_clamp, filename_E_2_jet_clamp_inv);
								
					error_code = -15;
					goto LABEL_ABEND;
				}
				else
				{
					printf("All E_2 files written successfully\n");
				
				}
			}/* E_2 calculated */


			/* isolines V_xy */
			
			/* isolines must be determined before writing to P3 file */
					
			if(0 !=	lbmn_anls_xy_D_create_all_isolines(
							ptr_isolines_V,/* output */
							ptr_V,/* input */
							ptr_Bnd,				
							size_row,
							size_col,
							V_min,
							V_max,
							number_of_isolines_V
				)
			)	
			{
				error_code = -16;
				goto LABEL_ABEND;
			}
			else
			{/* isolines V created */
			
				printf("All V_xy isolines created\n");
			
				if(0 != lbmn_anls_xy_D_fprintf_mesh_isolines_bw_netpbm_P3(
							filename_anls_V_xy_isolines_bw,
							ptr_Bnd,
							ptr_isolines_V,
							size_row,/* size of mesh */
							size_col,
							colour_isolines_background_bw_red ,/* colour - background */
							colour_isolines_background_bw_green,
							colour_isolines_background_bw_blue,
							ptr_isoline_colours_bw_red,/* colours of isolines */
							ptr_isoline_colours_bw_green,
							ptr_isoline_colours_bw_blue,
							number_of_isolines_V,
							ptr_electrode_colours_red,/* colours of electrodes */
							ptr_electrode_colours_green,
							ptr_electrode_colours_blue)
				)
				{
					error_code = -18;
					goto LABEL_ABEND;
				}
				else
				{
					printf("All V_xy isoline files written successfully\n");
				}
			
			}/* isolines V created */


			/* GOTO */
			/* if no problem, then to to OK*/
			goto LABEL_OK;

			/* else print diagnositic info*/
			LABEL_ABEND:
			
			printf("+++++++++++++\n");
			printf("+++ ERROR +++\n");
			printf("+++++++++++++\n");
			printf(">> could not finish data processing\n");
			printf(">> error code == %i\n", error_code);

			
		}/* Laplace solution has been found */
			
		
			
		/* continue processing */
		LABEL_OK:

		/* DESTROY allocated problem */

		if(problem_allocated != 0)
		{/* problem has beend dynamically allocated at the beginning of program */
			
			printf("deallocation of memory...");
	
			ret_val = lbmn_mesh_xy_problem_destroy(
						& size_row,/*output*/
						& size_col,
						& x_min,
						& x_max,
						& y_min,
						& y_max,			
						& ptr_V,
						& ptr_Bnd,
						& ptr_E_x,
						& ptr_E_y,
						& ptr_E_1,
						& ptr_E_2,
						& ptr_isolines_V,
						& ptr_isolines_E_x,
						& ptr_isolines_E_y,
						& ptr_isolines_E_1,
						& ptr_isolines_E_2,
						& ptr_workbench_colours_red,
						& ptr_workbench_colours_green,
						& ptr_workbench_colours_blue,
						& ptr_workbench_bw_red,
						& ptr_workbench_bw_green,
						& ptr_workbench_bw_blue);
		
			if(ret_val == 0)
			{
				printf("OK\n");
			}
			else
			{
				printf("FAILED: ret_val == %i\n", ret_val);
			}
		}/* problem has beend dynamically allocated at the beginning of program */
		
		
		printf("*** The end of DEMO_XY program ***\n");
		
    }/* problem to solve created */


    return error_code;

}/* main program */

/* ==== end of file demo_xy.c ==== */
