/* demo_zr.c
 * 
 * Copyright (C) 2024 Marcin Kulbaka
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>
#include "lbmn_relax_zr_D.h"

/**********************
* mesh geometry       *
* (size and z values) *
***********************/

#define NROW (100)
#define NCOL (600)
#define Z_02_MIN (0.0)
#define Z_02_MAX (300.0)

/****************
* visualization *
****************/

#define NCOLOURS (4)



/******************
* solver settings *
*******************/

#define ITER_MAX (1000000)
#define EPS (1.0e-9)


/*******************
* potential arrays *
********************/

double          tab_V   [NROW * NCOL];
unsigned int    tab_Bnd [NROW * NCOL];


double tab_z_02     	[NCOL];
double tab_z_1			[NCOL + 1]; /* for d1V_dz1*/

double tab_V_z     		[NCOL];

double tab_d1V_dz1_h 	[NCOL + 1]; /* slightly diferent size */
double tab_d2V_dz2_h	[NCOL];

double tab_d1V_dz1_2h 	[NCOL]; /* slightly diferent size */
double tab_d2V_dz2_2h	[NCOL];


/*******************
* grapher settings *
********************/

/* workbench - colours */
unsigned char tab_workbench_colours_red		[NROW * NCOL];
unsigned char tab_workbench_colours_green	[NROW * NCOL];
unsigned char tab_workbench_colours_blue	[NROW * NCOL];

/* workbench - bw	*/
unsigned char tab_workbench_bw_red		[NROW * NCOL];
unsigned char tab_workbench_bw_green	[NROW * NCOL];
unsigned char tab_workbench_bw_blue		[NROW * NCOL];


unsigned char tab_electrode_colours_red		[NCOLOURS];
unsigned char tab_electrode_colours_green	[NCOLOURS];
unsigned char tab_electrode_colours_blue	[NCOLOURS];



/************
* filenames *
*************/

const char* filename_V_zr_gray          		= "./outputs_zr/V_zr_gray.ppm";
const char* filename_V_zr_gray_inv      		= "./outputs_zr/V_zr_gray_inv.ppm";
const char* filename_V_zr_hottocold_clamp     	= "./outputs_zr/V_zr_hottocold.ppm";
const char* filename_V_zr_hottocold_clamp_inv 	= "./outputs_zr/V_zr_hottocold_inv.ppm";
const char* filename_V_zr_jet_clamp     		= "./outputs_zr/V_zr_jet.ppm";
const char* filename_V_zr_jet_clamp_inv 		= "./outputs_zr/V_zr_jet_inv.ppm";

const char* filename_Bnd_zr_dat         		= "./outputs_zr/Bnd_zr.dat";

/* files for V_z*/
const char* filename_V_zr_dat           		= "./outputs_zr/V_zr.dat";
const char* filename_V_z_dat 		    		= "./outputs_zr/V_z.dat";

/* files for V_z derivatives */
/* h */
const char* filename_d1V_dz1_h_dat      		= "./outputs_zr/V_d1_dz1_h.dat";
const char* filename_d2V_dz2_h_dat      		= "./outputs_zr/V_d2_dz2_h.dat";
/* 2h */
const char* filename_d1V_dz1_2h_dat     		= "./outputs_zr/V_d1_dz1_2h.dat";
const char* filename_d2V_dz2_2h_dat     		= "./outputs_zr/V_d2_dz2_2h.dat";

/* files for workbench visualization (electrodes) */

const char* filename_workbench_bw				="./outputs_zr/wbench_bw.ppm";
const char* filename_workbench_colours			="./outputs_zr/wbench_colours.ppm";


/****************** main program *********************/

int
main(void)
{
    double* ptr_V           = NULL;
    unsigned int* ptr_Bnd   = NULL;

	double* ptr_z_02        = NULL;
	double* ptr_z_1			= NULL;

    double* ptr_V_z         = NULL;
	
    /* h */
	double* ptr_d1V_dz1_h   = NULL;
    double* ptr_d2V_dz2_h   = NULL;

	/* 2h */
	double* ptr_d1V_dz1_2h  = NULL;
    double* ptr_d2V_dz2_2h  = NULL;

    unsigned int i_row = 0;
    unsigned int i_col = 0;
    
    int ret_val;
	

	int ret_val_V_z       = 0;
	
	int ret_val_d1V_dz1_h = 0;
	int ret_val_d2V_dz2_h = 0;

	int ret_val_d1V_dz1_2h = 0;
	int ret_val_d2V_dz2_2h = 0;	

    unsigned int boundary = 0;
    double voltage = 0.0;

    double tmp_min = 0.0;
    double tmp_max = 0.0;

    unsigned int i_iter = 0;
    unsigned int i_iter_max = ITER_MAX;
    
    double eps_temp = 0.0;

	unsigned int counter_1000 = 0;/* to display each 1000-th eps on display */


	/********************
	* geometry settigns *
	*********************/

	double z_02_min = Z_02_MIN; 
	double z_02_max = Z_02_MAX;

	double mesh_step_z = (z_02_max - z_02_min) / (NCOL - 1);


	/********************************************
	* workbench (BW) - colour of optical axis	*
	*********************************************/
	
	/* black */
	unsigned char axis_colour_red 	= 0;
	unsigned char axis_colour_green = 0;
	unsigned char axis_colour_blue	= 0;
	
	
	
    /********************************
    *                               *
    * initialize pointers to data   *
    *                               *
    ********************************/

    ptr_V       = tab_V;
    ptr_Bnd     = tab_Bnd;

	ptr_z_02 	= tab_z_02;
	ptr_z_1 	= tab_z_1;

    ptr_V_z     = tab_V_z;
	
    ptr_d1V_dz1_h  = tab_d1V_dz1_h;
    ptr_d2V_dz2_h  = tab_d2V_dz2_h;

	ptr_d1V_dz1_2h = tab_d1V_dz1_2h;
    ptr_d2V_dz2_2h = tab_d2V_dz2_2h;

	
	/********************************
	*								*
	* set workbench scenes to black	*
	*								*
	*********************************/
	
	for(i_col = 0; i_col < NROW*NCOL; i_col++)
	{
		tab_workbench_bw_red	[i_col] = 0;
		tab_workbench_bw_green	[i_col] = 0;
		tab_workbench_bw_blue	[i_col] = 0;
	}
	

	for(i_col = 0; i_col < NROW*NCOL; i_col++)
	{
		tab_workbench_colours_red	[i_col] = 0;
		tab_workbench_colours_green	[i_col] = 0;
		tab_workbench_colours_blue	[i_col] = 0;
	}

	/********************************
	* set colours for electrodes	*
	********************************/
	
	/* 0 - vacuum -> white	*/
	tab_electrode_colours_red	[0] = 255;
	tab_electrode_colours_green	[0] = 255;
	tab_electrode_colours_blue	[0] = 255;
	
	/* 1 -> red 			*/
	tab_electrode_colours_red	[1] = 128;
	tab_electrode_colours_green	[1] = 0;
	tab_electrode_colours_blue	[1] = 0;
	
	/* 2 -> green			*/
	tab_electrode_colours_red	[2] = 0;
	tab_electrode_colours_green	[2] = 128;
	tab_electrode_colours_blue	[2] = 0;


	/* 3 -> blue			*/
	tab_electrode_colours_red	[3] = 0;
	tab_electrode_colours_green	[3] = 0;
	tab_electrode_colours_blue	[3] = 128;

	


    /********************
    * clear tables      *
    * voltage = 0.0 V   *
    * boundary = 0      *
    *********************/
    voltage = 0.0;
    boundary = 0;

    for(i_row = 1; i_row <= NROW; i_row++)
    {

        for(i_col = 1; i_col <= NCOL; i_col++)
        {
            ret_val = lbmn_relax_zr_D_set_point(
                            ptr_V,
                            ptr_Bnd,
                            NROW,/* size of mesh */
                            NCOL,
                            i_row,/* node position (1,2, .. , n_row)*/
                            i_col,/* 1,2,..,n_col*/
                            voltage, /*input node values*/
                            boundary);


            if(ret_val != 0)
            {
                printf("FAILED - could not initialize mesh correctly\n");
                break;
            }

        }/*i_col*/

    }/*i_row*/


	/************************
	* clear tables:			*
	*						*
	*   V(z)				*
	*   V'(z)				*
	*   V''(z)  			*
	************************/
	
	for(i_col = 1; i_col <= NCOL; i_col++)
	{

		*(ptr_V_z     + i_col - 1)   = 0.0;
		
		*(ptr_d1V_dz1_h + i_col - 1) = 0.0;
		*(ptr_d2V_dz2_h + i_col - 1) = 0.0;
		
		*(ptr_d1V_dz1_2h + i_col - 1) = 0.0;
		*(ptr_d2V_dz2_2h + i_col - 1) = 0.0;

	}
	
	*(ptr_d1V_dz1_h + NCOL) = 0.0; /* last element of d1V_dz1*/


	/* all tables have been filled with zeros */	
	
	
	/****************************
	* init z tables
	*
	*
	*
	*****************************/

	for(i_col = 1; i_col <= NCOL; i_col++)
	{
		*(ptr_z_02 + i_col - 1) =  z_02_min  + (double)(i_col-1) * mesh_step_z;
		
		*(ptr_z_1  + i_col - 1) =  *(ptr_z_02 + i_col - 1) - 0.5 * mesh_step_z;

	}
	/* all tables have been filled with zeros */	
	/* initialize z points*/


	/* last point has been set */
	*(ptr_z_1 + NCOL ) = z_02_min + (double)NCOL * mesh_step_z;
	






    printf("+------------------------------+\n");
    printf("|        RELAX Z-R (D)         |\n");
    printf("+------------------------------+\n");



    /****************************
    *                           *
    * setting electrode points  *
    *                           *
    ****************************/

/*
* electrode 1
*/
    boundary    = +1;
    voltage     = +0.0;

	for(i_row = 91; i_row <= 100; i_row++)
	{
		for(i_col = 1; i_col <= 200; i_col++)
		{
			ret_val = lbmn_relax_zr_D_set_point(
                            ptr_V,
                            ptr_Bnd,
                            NROW,/* size of mesh */
                            NCOL,
                            i_row,/* node position (1,2, .. , n_row)*/
                            i_col,/* 1,2,..,n_col*/
                            voltage, /*input node values*/
                            boundary);
		}
    }

/*
* electrode 2
*/
    boundary    = +2;
    voltage     = -1.0;

	for(i_row = 91; i_row <= 100; i_row++)
	{
		for(i_col = 201; i_col <= 400; i_col++)
		{
			ret_val = lbmn_relax_zr_D_set_point(
                            ptr_V,
                            ptr_Bnd,
                            NROW,/* size of mesh */
                            NCOL,
                            i_row,/* node position (1,2, .. , n_row)*/
                            i_col,/* 1,2,..,n_col*/
                            voltage, /*input node values*/
                            boundary);
		}
    }
  

/*
* electrode 3
*/
    boundary    = +3;
    voltage     = +0.0;

	for(i_row = 91; i_row <= 100; i_row++)
	{
		for(i_col = 401; i_col <= 600; i_col++)
		{
			ret_val = lbmn_relax_zr_D_set_point(
                            ptr_V,
                            ptr_Bnd,
                            NROW,/* size of mesh */
                            NCOL,
                            i_row,/* node position (1,2, .. , n_row)*/
                            i_col,/* 1,2,..,n_col*/
                            voltage, /*input node values*/
                            boundary);
		}
    }

   
    /*
    * displaying electrode points
    */

    /*
     * 1 / 2 - finding min / max
    */




    /*
    * part 2 / 2
    */
    printf("\n");
    printf("V on mesh:\n\n");
    lbmn_relax_zr_D_printf_V_zr(
                        ptr_V,
                        NROW,
                        NCOL);

    printf("\n");
    printf("boundaries on mesh:\n\n");
    lbmn_relax_zr_D_printf_Bnd_zr(
                        ptr_Bnd,
                        NROW,
                        NCOL);

    

    /********************
    *                   *
    * solve Laplace     *
    *                   *
    *********************/

    i_iter = 1;
	counter_1000 = 1;

    eps_temp = lbmn_relax_zr_D_iter(ptr_V, ptr_Bnd, NROW, NCOL);

    printf("\n>>> Solving \n\n");

    printf("%7d ---  %+18.15E\n", i_iter, eps_temp);

    while(
            (eps_temp > EPS)
            &&
            (i_iter < i_iter_max)
         )
    {
        eps_temp = lbmn_relax_zr_D_iter(ptr_V, ptr_Bnd, NROW, NCOL);

        i_iter++;

		counter_1000 ++;
        		
		if(counter_1000 >= 1000)
		{
			counter_1000 = 0; /*reset counter*/
			
			/*display another 1000-th eps */
			printf("%7d ---  %+18.15E\n", i_iter, eps_temp);
		}
    }

    if(i_iter < i_iter_max)
    {    

        /* === end of reraxation, solution found (+/-)!!!  === */

        printf("\n\n");

        printf(">>> SUCCESS - solution has been found\n");
        
        printf("after %d iterations\n\n", i_iter);

        printf("Searching for min / max\n\n");

        lbmn_relax_zr_D_find_V_min_max(
                            &tmp_min,
                            &tmp_max,
                            ptr_V,  
                            NROW,
                            NCOL);


        printf("V_min == %+f\n", tmp_min);
        printf("V_max == %+f\n", tmp_max);

        /*
        * printf show solved V / Bnd
        */


/*
        printf("\n");
        printf("V on mesh:\n\n");
        RELAX_ZR_printf_V_zr(ptr_V,
                            NROW,
                            NCOL);


        printf("\n");
        printf("boundaries on mesh:\n\n");
        RELAX_ZR_printf_Bnd_zr(ptr_Bnd,
                            NROW,
                            NCOL);
*/

		/****************************
		*							*
		* create workbench image	*
		*							*
		*****************************/
	
		if(lbmn_relax_zr_D_create_image_workbench_bw(
				tab_workbench_bw_red,
				tab_workbench_bw_green,
				tab_workbench_bw_blue,
				NROW,/* size of mesh */
				NCOL,
				ptr_Bnd,/* mesh of boundaries*/
				axis_colour_red,
				axis_colour_green,
				axis_colour_blue)
			||
			
			lbmn_relax_zr_D_create_image_workbench_colours(
				tab_workbench_colours_red,
				tab_workbench_colours_green,
				tab_workbench_colours_blue,
				NROW,/* size of mesh */
				NCOL,
				ptr_Bnd,/* mesh of boundaries*/
				tab_electrode_colours_red,
				tab_electrode_colours_green,
				tab_electrode_colours_blue,
				NCOLOURS,
				axis_colour_red,
				axis_colour_green,
				axis_colour_blue)
				)
			
		{/* problem with creation of workbench */
		
			printf("ERROR - could not create all workbench images\n");
			
			
		}/* problem with creation of workbench */			
		else
		{/* success */
	
			printf("workbench images (BW/colour) created successfully\n");
			
		}/* success */
			
			
        if(
				lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_gray,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_grayscale)
			||	
				lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_gray_inv,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_grayscale_inv)
            || 
                lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_hottocold_clamp,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_hottocold_clamp)
			||
				lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_hottocold_clamp_inv,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_hottocold_clamp_inv)
			||
				lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_jet_clamp,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_jet_clamp)
			||
				lbmn_relax_zr_D_fprintf_V_netpbm_P3(
                filename_V_zr_jet_clamp_inv,
                ptr_V,
                NROW,/* size of mesh */
                NCOL,
                lbmn_colormap_jet_clamp_inv)
			||
				lbmn_relax_zr_D_fprintf_any_workbench_netpbm_P3(
				filename_workbench_bw,
				tab_workbench_bw_red,
				tab_workbench_bw_green,
				tab_workbench_bw_blue,
				NROW,/* size of mesh */
				NCOL)
			||
				lbmn_relax_zr_D_fprintf_any_workbench_netpbm_P3(
				filename_workbench_colours,
				tab_workbench_colours_red,
				tab_workbench_colours_green,
				tab_workbench_colours_blue,
				NROW,/* size of mesh */
				NCOL)
			
		)
        {
            printf("ERROR - could not write PPM P3 data to file(s)\n%s\n%s\n%s\n%s\n%s\n%s\n",
						filename_V_zr_gray, filename_V_zr_gray_inv,
						filename_V_zr_hottocold_clamp, filename_V_zr_hottocold_clamp_inv,
						filename_V_zr_jet_clamp, filename_V_zr_jet_clamp_inv);
        }
        else
        {
            printf("PPM P3 file(s) written successfully\n");
        }
        


        if(lbmn_relax_zr_D_fprintf_any_V_z(
                                filename_V_z_dat,
                                ptr_z_02,
                                ptr_V_z, 
                                NCOL))
        {
            printf("ERROR - could not write V_z to file=%s\n", filename_V_z_dat);
        }
        else
        {
            printf("V_z file written successfully\n");
        }


        if(lbmn_relax_zr_D_fprintf_V_dat(
                    filename_V_zr_dat,
                    ptr_V,
                    NROW,/* size of mesh */
                    NCOL)
        )
        {
             printf("ERROR - could not write V_zr to file %s\n", filename_V_zr_dat);
        }
        else
        {
            printf("V_zr  file written successfully\n");
        }

           

        if(lbmn_relax_zr_D_fprintf_Bnd_dat(
                            filename_Bnd_zr_dat,
                            ptr_Bnd,
                            NROW,/* size of mesh */
                            NCOL)
        )
        {
            printf("ERROR - could not write Bnd_zr to file=%s\n", filename_Bnd_zr_dat);
        }
        else
        {
            printf("Bnd_zr file written successfully\n");
        }


		/****************************
		* calculate 				*
		*							*
		* V_z						*
		* d1V_dz1					*
		* d2V_dz2					*
		*							*
		*****************************/
		
		if(lbmn_relax_zr_D_calculate_all_tab_V_z_h(
					ptr_V_z,
					ptr_d1V_dz1_h,
					ptr_d2V_dz2_h,
					ptr_V,
					NCOL,
					mesh_step_z)
			||
			lbmn_relax_zr_D_calculate_tab_V_z_2h(
					ptr_d1V_dz1_2h,
					ptr_d2V_dz2_2h,
					ptr_V,
					NCOL,
					mesh_step_z)
					
		)
		{/* error */
			
			printf("ERROR - could not caclulate all_tab_V_z (h or 2h)");
			
		}/* error */
		else
		{/* tables V_z calculated */
	
			printf("tables V_z calculated\nwriting data to files...\n");
			
			ret_val_V_z       = lbmn_relax_zr_D_fprintf_any_V_z(
                                                filename_V_z_dat,
                                                ptr_z_02,
                                             	ptr_V_z,
                                        		NCOL);
			
			ret_val_d1V_dz1_h = lbmn_relax_zr_D_fprintf_any_V_z(
                                                filename_d1V_dz1_h_dat,
                                                ptr_z_1,
                                                ptr_d1V_dz1_h,
                                                NCOL + 1);

			ret_val_d2V_dz2_h = lbmn_relax_zr_D_fprintf_any_V_z(
                                                filename_d2V_dz2_h_dat,
                                                ptr_z_02,
                                                ptr_d2V_dz2_h,
                                                NCOL);
			
			ret_val_d1V_dz1_2h = lbmn_relax_zr_D_fprintf_any_V_z(
                                                filename_d1V_dz1_2h_dat,
                                                ptr_z_02,
                                                ptr_d1V_dz1_2h,
                                                NCOL);

			ret_val_d2V_dz2_2h = lbmn_relax_zr_D_fprintf_any_V_z(
                                                filename_d2V_dz2_2h_dat,
                                                ptr_z_02,
                                                ptr_d2V_dz2_2h,
                                                NCOL);

			if( (ret_val_V_z 		!= 0) &&
				(ret_val_d1V_dz1_h 	!= 0) &&
				(ret_val_d2V_dz2_h 	!= 0) &&
				(ret_val_d1V_dz1_2h != 0) &&
				(ret_val_d2V_dz2_2h	!= 0)
			)
			{
				printf("ERROR - could not write all V_z tables to files\n");
			}
			else
			{
				printf("Success. All files V_z has been written to hard disk\n");
			}
									
		}/* tables V_z calculated */
			

    }/* solution found */
    else
    {
        /* given accuracy nas not been achieved */

        printf("ERROR!!!\n");
        printf("LAPLACE SOLUTION HAS NOT BEEN FOUND FOR GIVEN PARAMETERS!!!\n ");
    }

    return 0;

}/* main */

/* ==== end of file demo_zr.c ==== */