/* lbmn_relax_xy_D.c
 * 
 * Copyright (C) 2024 Marcin Kulbaka
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "lbmn_relax_xy_D.h"

/* single relaxation procedure */
double
lbmn_relax_xy_D_iter(
         double*        ptr_V,
         unsigned int*  ptr_Bnd,
         unsigned int   size_row,/* size of mesh */
         unsigned int   size_col)
{
	/* declaration of variables */
	
    double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

	unsigned int i_row = 0; /* indices used in loops */
    unsigned int i_col = 0;    


	/* instructions*/
	
 	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh
	*
	*	xyLV_RELAX5_P1_D .. xyLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (XY)
	* (CARTETSIAN COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/

	/*
	* mesh processing scheme (the same as BR):
	*
	* P1 - P2 loop - P3 (bottom row)
	* P4 - P5 loop - P6 (inner rows)
	* P7 - P8 loop - P9 (upper row)
	*/
	
    /* === P1 === */

	/* new value of icp */
    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P1_D */
		
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P1, if Bnd == 0 */


    /* === P2 loop === */

	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

		/* === P2 === */
		
        /*new value of icp*/
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P2_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P2,  if Bnd == 0*/

    }/* P2 loop (i_col)*/


    /* === P3 === */

    /* new value of icp */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);
		
		/* relaxation xyLV_RELAX5_P3_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) 
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P3, if Bnd == 0 */


    /* === P4-P5-P6 loop === */
	
	/* direction: bottom -> up */
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {
        /* === P4 === */

        /* new value of icp */
        icp = (i_row - 1) * size_col;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);
			
			/* relaxation xyLV_RELAX5_P4_D */
            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) 
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0 */


        /*=== P5 loop ===*/
		
        /* direction: left -> right */
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {
			
			/* === P5 === */
            /* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

				/* relaxation xyLV_RELAX5_P5_D */
                *(ptr_V + icp) =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        +
                        *(ptr_V + icp + size_col)
                        +
                        *(ptr_V + icp - size_col)
                        )
                        *
                        0.25;       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/*P5 , if Bnd == 0*/

        }/* P5 loop (i_col)*/


        /* === P6 === */

        /* new value of icp */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P6_D */
            *(ptr_V + icp)
                        =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

    }/* loop P4 - P5 - P6 */


    /* === P7 === */

	/* new value of icp */
    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P7_D */
        *(ptr_V + icp) =
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/


    /*=== P8 loop ===*/

	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

		/* === P8 === */
		
        /*new value of icp*/
        icp = (size_row - 1) * size_col + i_col - 1;
		

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);
			
			/* relaxation xyLV_RELAX5_P8_D */
            *(ptr_V + icp) = 
                        (
                        *(ptr_V + icp - size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P8, if Bnd == 0*/

    }/* P8 loop */


    /* === P9 === */

    /* new value of icp */
    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);
		
		/* relaxation xyLV_RELAX5_P9_D */
        *(ptr_V + icp) = 
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp - 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P9, if Bnd == 0*/


    /* end */
    return eps_max;
	
}/* lbmn_relax_xy_D_iter */

/*================================================================================*/

double
lbmn_relax_xy_D_iter_BL(
           double*          ptr_V,
           unsigned int*    ptr_Bnd,
           unsigned int     size_row,/* size of mesh */
           unsigned int     size_col)
{
	
	/* declaration of variables */
	
	double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    unsigned int i_row = 0;	/*indices used in loops*/
    unsigned int i_col = 0;    


	/* instructions */

	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh	
	*
	*	xyLV_RELAX5_P1_D .. xyLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (XY)
	* (CARTETSIAN COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/
	
	/*
	* mesh processing scheme (BL):
	*
	* P3 - P2 loop - P1 (bottom row)
	* P6 - P5 loop - P4 (inner rows)
	* P9 - P8 loop - P7 (upper row)
	*/
	
	
	/* === P3 === */

	/* new value of icp */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P3_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) 
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P3, if Bnd == 0*/
	
    /*=== loop P2 ===*/
	
	/* direction: right -> left */
    for(i_col = size_col - 1; i_col >= 2; i_col--)
    {

		/* new value of icp */
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P2_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* loop P2 */

    /* === P1 === */

	/* new value of icp */
    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P1_D */
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P1, if Bnd == 0*/
  

    /*=== loop P6-P5-P4 ===*/
	
	/* direction: bottom -> up */
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {

        /* === P6 === */

        /* new value of icp */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P6_D */
            *(ptr_V + icp)
                        =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/


        /* === loop P5 ===*/
		
		/* direction: right -> left */
        for(i_col = size_col - 1; i_col >= 2; i_col--)
        {
			
            /* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {
                tmp = *(ptr_V + icp);

				/* relaxation xyLV_RELAX5_P5_D */
                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        +
                        *(ptr_V + icp + size_col)
                        +
                        *(ptr_V + icp - size_col)
                        )
                        *
                        0.25;       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* loop P5 */


		/* === P4 ===*/
		
        /* new value of icp */
        icp = (i_row - 1) * size_col;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P4_D */
            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) 
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/

    }/* loop P6-P5-P4 */



    /* === P9 === */

    /* new value of icp */
    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P9_D */
        *(ptr_V + icp) = 
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp - 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P9, if Bnd == 0*/

  
    /* === P8 loop === */

	/* direction: right -> left */
    for(i_col = size_col - 1; i_col >= 2; i_col--)
    {

		/* === P8 === */
		
        /*new value of icp*/
        icp = (size_row - 1) * size_col + i_col - 1;
    
        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P8_D */
            *(ptr_V + icp) = 
                        (
                        *(ptr_V + icp - size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P8, if Bnd == 0*/

    }/*P8 loop */



    /* === P7 === */

	/* new value of icp */
    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P7_D */
        *(ptr_V + icp) =
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/

    /* end */
    return eps_max;
	
}/* lbmn_relax_xy_D_iter_BL */
		   
/*================================================================================*/
		   
double
lbmn_relax_xy_D_iter_BR(
            double*         ptr_V,
            unsigned int*   ptr_Bnd,
            unsigned int    size_row,/* size of mesh */
            unsigned int    size_col)
{
	/* declaration of variables */
	
	double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    unsigned int i_row = 0;/* indices used in loops */
    unsigned int i_col = 0; 

	/* instructions */
	
	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh	
	*
	*	xyLV_RELAX5_P1_D .. xyLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (XY)
	* (CARTETSIAN COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/	
	
	/*
	* mesh processing scheme (BR):
	*
	* P1 - P2 loop - P3 (bottom row)
	* P4 - P5 loop - P6 (inner rows)
	* P7 - P8 loop - P9 (upper row)
	*/

    /* === P1 === */

	/* new value of icp */
    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P1_D */
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P1, if Bnd == 0*/


    /* === P2 loop === */
	
	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {
		
        /*new value of icp*/
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P2_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* P2 loop */


    /* === P3 === */

	/* new value of icp */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P3_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) 
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P3, if Bnd == 0*/

  
    /*=== loop P4-P5-P6 ===*/

	/* direction: bottom -> up */
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {

		/* === P4 === */
		
		/* new value of icp */
        icp = (i_row - 1) * size_col;
        
        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P4_D */
            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) 
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/


        /* === P5 loop === */

		/* direction: left -> right */		
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {

			/* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            /* P5 */
            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

				/* relaxation xyLV_RELAX5_P5_D */
                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        +
                        *(ptr_V + icp + size_col)
                        +
                        *(ptr_V + icp - size_col)
                        )
                        *
                        0.25;       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* P5 loop */

        /* === P6 === */

		/* new value of icp */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P6_D */
            *(ptr_V + icp)
                        =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

    }/* loop P4 - P5 - P6 */


    /* === P7 === */

	/* new value of icp */
    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P7_D */
        *(ptr_V + icp) =
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/

  
    /* === loop P8 === */

	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

		/* === P8 === */
		
        /* new value of icp */
        icp = (size_row - 1) * size_col + i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P8_D */
            *(ptr_V + icp) = 
                        (
                        *(ptr_V + icp - size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P8, if Bnd == 0*/

    }/*P8 loop */


    /* === P9 === */

	/* new value of icp */
    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P9_D */
        *(ptr_V + icp) = 
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp - 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P9, if Bnd == 0*/


    /* end */
    return eps_max;

}/* lbmn_relax_xy_iter_D_BR */


/*================================================================================*/


double
lbmn_relax_xy_D_iter_TL(
            double*         ptr_V,
            unsigned int*   ptr_Bnd,
            unsigned int    size_row,/* size of mesh */
            unsigned int    size_col)
{
	/* declaration of variables */
	
	double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    unsigned int i_row = 0;	/*indices used in loops*/
    unsigned int i_col = 0;    


	/* instructions */

	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh	
	*
	*	xyLV_RELAX5_P1_D .. xyLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (XY)
	* (CARTETSIAN COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/
	
	/*
	* mesh processing scheme (TL):
	*
	* P9 - P8 loop - P7 (upper row)
	* P6 - P5 loop - P4 (inner rows)
	* P3 - P2 loop - P1 (bottom row)
	*/
	
	
	/* === P9 === */

    /* new icp value */
    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P9_D */
        *(ptr_V + icp) = 
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp - 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P9, if Bnd == 0*/

  
    /*=== P8 loop ===*/

	/* direction: right -> left */
    for(i_col = size_col - 1; i_col >= 2; i_col--)
    {

		/* new value of icp */
        icp = (size_row - 1) * size_col + i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P8_D */
            *(ptr_V + icp) = 
                        (
                        *(ptr_V + icp - size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P8, if Bnd == 0*/

    }/*P8 loop */


    /* === P7 === */

	/* new value of icp */
    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P7_D */
        *(ptr_V + icp) =
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/
	

    /*=== loop P6-P5-P4, Top-Down ===*/
	
	/* direction: top -> down */
    for(i_row = size_row - 1 ; i_row >= 2; i_row-- )
    {

        /* === P6 === */

        /* new value of icp */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P6_D */
            *(ptr_V + icp)
                        =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

        /* === P5 loop ===*/

		/* direction: right -> left */		
        for(i_col = size_col - 1; i_col <= 2; i_col--)
        {
			
            /* === P5 ===*/
			
            /* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

				/* relaxation xyLV_RELAX5_P5_D */
                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        +
                        *(ptr_V + icp + size_col)
                        +
                        *(ptr_V + icp - size_col)
                        )
                        *
                        0.25;       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* P5 loop (i_col) */

        
		/* === P4 === */
		
		/* new value of icp */
        icp = (i_row - 1) * size_col;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P4_D */
            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) 
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/

    }/* loop P6-P5-P4 */


	/* === P3 === */

	/* new value of icp */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P3_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) 
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P3, if Bnd == 0*/
	
	
    /*=== P2 loop ===*/
	
	/* direction: right -> left */
    for(i_col = size_col - 1; i_col >= 2; i_col--)
    {

		/*new value of icp*/
        icp = i_col - 1;


        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P2_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* P2 loop */



    /* === P1 === */

	/* new value of icp */
    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P1_D */
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P1, if Bnd == 0*/


    /* end */
    return eps_max;

}/* lbmn_relax_xy_D_iter_TL */


/*================================================================================*/


double
lbmn_relax_xy_D_iter_TR(
            double*         ptr_V,
            unsigned int*   ptr_Bnd,
            unsigned int    size_row,/* size of mesh */
            unsigned int    size_col)
{
	/* declaration of variables */
	
	double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    unsigned int i_row = 0;	/*indices used in loops*/
    unsigned int i_col = 0;    


	/* instructions */

	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh	
	*
	*	xyLV_RELAX5_P1_D .. xyLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (XY)
	* (CARTETSIAN COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/
	
	/*
	* mesh processing scheme (TR):
	*
	* P7 - P8 loop - P9 (upper row)
	* P4 - P5 loop - P6 (inner rows)
	* P1 - P2 loop - P3 (bottom row)
	*/
	
	
    /* === P7 === */

	/* new value of icp */
    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P7_D */
        *(ptr_V + icp) =
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/

  
    /*=== P8 loop ===*/
	
	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++ )
    {

		/*new value of icp*/
        icp = (size_row - 1) * size_col + i_col - 1;
 
 
        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P8_D */
            *(ptr_V + icp) = 
                        (
                        *(ptr_V + icp - size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P8, if Bnd == 0*/

    }/*P8 loop */


	/*=== P9 === */

    /* new icp value */
    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P9_D */
        *(ptr_V + icp) = 
                    (
                    *(ptr_V + icp - size_col)
                    +
                    *(ptr_V + icp - 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P9, if Bnd == 0*/
  

    /*=== loops P4-P5-P6 ===*/
	
	/* direction: top -> down */
    for(i_row = size_row - 1; i_row >= 2; i_row--)
    {

		/* === P4 === */
		
        /* new value of icp */
        icp = (i_row - 1) * size_col;
        
        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P4_D */
            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) 
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/



        /* === P5 loop ===*/
         
		/* direction: left -> right */
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {

			/* === P5 === */
			
            /* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P5_D */
                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        +
                        *(ptr_V + icp + size_col)
                        +
                        *(ptr_V + icp - size_col)
                        )
                        *
                        0.25;       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* P5 loop (i_col) */


       /*=== P6 ===*/

        /* new value of icp */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P6_D */
            *(ptr_V + icp)
                        =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp - size_col)
                        +
                        *(ptr_V + icp + size_col)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/
 
    }/* loop P4-P5-P6 */


    /*=== P1 ===*/

	/* new value of icp */
    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P1_D */
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col)
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P1, if Bnd == 0*/

	
    /*=== P2 loop ===*/

	/* direction: left -> right */
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of icp*/
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation xyLV_RELAX5_P2_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col)  
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        3.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* P2 loop */


	/*=== P3 ===*/

	/* new value of icp */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation xyLV_RELAX5_P3_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) 
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.5;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P3, if Bnd == 0*/
	
    /* end */
    return eps_max;

}/* lbmn_relax_xy_D_iter_TR */


/* ==== end of file lbmn_relax_xy_D.c ==== */
