/* lbmn_relax_zr_D.c
 * 
 * Copyright (C) 2024 Marcin Kulbaka
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "lbmn_relax_zr_D.h"


/* single relaxation procedure */
double
lbmn_relax_zr_D_iter(
        double*         ptr_V,
        unsigned int*   ptr_Bnd,
        unsigned int    size_row,/* size of mesh */
        unsigned int    size_col)
{
	/* declaration of variables */
	
    double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    double alpha = 0.0;
    double beta  = 0.0;

    unsigned int i_row = 0;
    unsigned int i_col = 0;    

	/* instructions */

	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh
	*
	*	zrLV_RELAX5_P1_ON_D .. zrLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (ZR)
	* (CYLINDRICAL COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/

	/*
	* mesh processing scheme (the same as BR):
	*
	* P1 - P2 loop - P3 (bottom row)
	* P4 - P5 loop - P6 (inner rows)
	* P7 - P8 loop - P9 (upper row)
	*/
	
    /*=== P1 ===*/

    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation zrLV_RELAX5_P1_ON_D */		
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col) * 4.0
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P1, if Bnd == 0*/


    /*=== loop P2 ===*/
	
	/* direction: left -> right */   
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of ipc*/
        icp = i_col - 1;


        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation zrLV_RELAX5_P2_ON_D */
			*(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col) * 4.0 
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        6.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/*=== loop P2 ===*/



    /*=== P3 ===*/

    /* new icp value */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation zrLV_RELAX5_P3_ON_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) * 4.0
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P3, if Bnd == 0*/


    /*=== loop P4-P5-P6 === */

	/* direction: bottom -> up */   
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {

		/*=== P4 ===*/

        /* new icp value */
        icp = (i_row - 1) * size_col;

        	
        if(*(ptr_Bnd + icp) == 0)
        {

            alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1) / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) / 3.0
                        )
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/


        /*=== loop P5 ===*/
		
        /* new values for each row */
        alpha = (double)(2 * i_row - 3) /  (double)(8 * i_row - 8);
        beta  = (double)(2 * i_row - 1) /  (double)(8 * i_row - 8);
         

		/* direction: left -> right */        
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {
            
			/* new value of icp */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        *
                        0.25
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* loop P5 */


        /*=== P6 ===*/

        /* new icp value */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            alpha = (double)(2 * i_row - 3)  / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1)  / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                        =
                        *(ptr_V + icp - 1)
                        /
                        3.0
                        +
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        +
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        ;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

    }/*=== loop P4 - P5 - P6 ===*/


    /*=== P7 ===*/

    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
        beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);

        tmp = *(ptr_V + icp);

        *(ptr_V + icp) =
                    alpha
                    *
                    (
                    *(ptr_V + icp - size_col)
                    )
                    +
                    beta
                    *
                    (
                    *(ptr_V + icp + 1)
                    );

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/


    /*=== loop P8 ===*/

	/* direction: left -> right */   
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of ipc*/

        alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 7); 
        beta  = (double)(2 * i_row - 2) / (double)(6 * i_row - 7);

        icp = (size_row - 1) * size_col + i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

            *(ptr_V + icp) = 
                        alpha
                        *
                        *(ptr_V + icp - size_col)  
                        +
                        beta
                        *         
                        (
                        *(ptr_V + icp - 1) 
                        +  
                        *(ptr_V + icp + 1)
                        );

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/*=== loop P8 ===*/


    /*=== P9 ===*/

    /* new icp value */

    i_row = size_row;

	icp = size_row * size_col - 1;
 
	alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
    beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);
    
    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

        *(ptr_V + icp) = 
                    alpha
                    *
                    *(ptr_V + icp - size_col)
                    +
                    beta
                    *
                    *(ptr_V + icp - 1);

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P9, if Bnd == 0*/


    /* end */
    return eps_max;
	
}/* lbmn_relax_zr_D_iter */

/*================================================================================*/

/* single relaxation procedure */
double
lbmn_relax_zr_D_iter_BL(
        double*   ptr_V,
        unsigned int*  ptr_Bnd,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col)
{
	/* declaration of variables */
	
    double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    double alpha = 0.0;
    double beta  = 0.0;

    unsigned int i_row = 0;
    unsigned int i_col = 0;    

	/* instructions */
	
	/* !!!!!!!!!!!!!!!! TODO !!!!!!!!!!!!!!!!!!!!! */
	/*
	P3 - loop P2 - P1 - this is done
	P6 - loop P5 - P4 - TODO !
	P9 - loop P8 - P7 - TODO !
	*/
	
	
	/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh
	*
	*	zrLV_RELAX5_P1_ON_D .. zrLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (ZR)
	* (CYLINDRICAL COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/

	/*
	* mesh processing scheme (the same as BR):
	*
	* P3 - P2 loop - P1 (bottom row)
	* P6 - P5 loop - P4 (inner rows)
	* P9 - P8 loop - P7 (upper row)
	*/
	
	
    /*=== P3 ===*/

    /* new icp value */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation: zrLV_RELAX5_P3_ON_D*/
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) * 4.0
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P3, if Bnd == 0*/


    /*=== loop P2 ===*/

	/* direction: right -> left */
    for(i_col = size_col - 1; i_col >= 2; i_col--)
    {

        /* new value of icp */
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation: zrLV_RELAX5_P2_ON_D*/
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col) * 4.0 
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        6.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* loop  P2 */


    /*=== P1 ===*/

    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation: zrLV_RELAX5_P1_ON_D*/
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col) * 4.0
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P1, if Bnd == 0*/
	
	/*
	TODO 
	TODO
	*/
    /*=== loop P4-P5-P6 === */

    /*=== P4 ===*/
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {
        /* new ipc value */
        icp = (i_row - 1) * size_col;

        if(*(ptr_Bnd + icp) == 0)
        {

            alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1) / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) / 3.0
                        )
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/


        /*=== loop P5 ===*/

        /* new values for each row */
        alpha = (double)(2 * i_row - 3) /  (double)(8 * i_row - 8);
        beta  = (double)(2 * i_row - 1) /  (double)(8 * i_row - 8);
         
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {
            
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {

                tmp = *(ptr_V + icp);

                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        *
                        0.25
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* loop P5 */

        /*=== P6 ===*/

        /* new icp value */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {

            alpha = (double)(2 * i_row - 3)  / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1)  / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                        =
                        *(ptr_V + icp - 1)
                        /
                        3.0
                        +
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        +
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        ;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

    }/* loop P4 - P5 - P6 */


    /*=== P7 ===*/

    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {
        alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
        beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);

        tmp = *(ptr_V + icp);

        *(ptr_V + icp) =
                    alpha
                    *
                    (
                    *(ptr_V + icp - size_col)
                    )
                    +
                    beta
                    *
                    (
                    *(ptr_V + icp + 1)
                    );

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/


    /*=== loop P8 ===*/

    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of icp*/
        icp = (size_row - 1) * size_col + i_col - 1;


        alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 7);
        beta  = (double)(2 * i_row - 2) / (double)(6 * i_row - 7);

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

            *(ptr_V + icp) = 
                        alpha
                        *
                        *(ptr_V + icp - size_col)  
                        +
                        beta
                        *         
                        (
                        *(ptr_V + icp - 1) 
                        +  
                        *(ptr_V + icp + 1)
                        );

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* loop P8 */



    /*=== P9 ===*/

    /* new ipc value */


    icp = size_row * size_col - 1;

    i_row = size_row;

    alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
    beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

        *(ptr_V + icp) = 
                    alpha
                    *
                    *(ptr_V + icp - size_col)
                    +
                    beta
                    *
                    *(ptr_V + icp - 1);

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P9, if Bnd == 0*/


    /* end */
    return eps_max;
	
}/* lbmn_relax_zr_D_iter_BL */


/*================================================================================*/

/* single relaxation procedure */
double
lbmn_relax_zr_D_iter_BR(
        double*   ptr_V,
        unsigned int*  ptr_Bnd,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col)
{
	/* declaration of variables */
	
    double eps = 0.0;
    double eps_max = 0.0;
    double tmp = 0.0;
    
    unsigned int icp = 0;/* index of 'central point' for each case P1,..,P9 */

    double alpha = 0.0;
    double beta  = 0.0;

    unsigned int i_row = 0;
    unsigned int i_col = 0;    

	/* instructions */
	
		/*
	* let us apply relaxation precedure 
	* over all the non - vacuum nodes of mesh
	*
	*	zrLV_RELAX5_P1_ON_D .. zrLV_RELAX5_P9_D
	*
	* see:
	*
	* LIEBMANN TECHNICAL DOCUMENTATION
	* LAPLACE EQUATION 2D (ZR)
	* (CYLINDRICAL COORDIANATES)
	* RELAXATION SCHEME EXPLAINED
	* (5 - POINT STAR))
	*
	*/

	/*
	* mesh processing scheme:
	*
	* P1 - P2 loop - P3 (bottom row)
	* P4 - P5 loop - P6 (inner rows)
	* P7 - P8 loop - P9 (upper row)
	*/
	
    /*=== P1 ===*/

    icp = 0;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation zrLV_RELAX5_P1_ON_D */
        *(ptr_V + icp) =
                   (*(ptr_V + icp + size_col) * 4.0
                    +
                    *(ptr_V + icp + 1)
                    )
                    *
                    0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P1, if Bnd == 0*/

    /*=== loop P2 ===*/

	/* direction: left -> right */   
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of ipc*/
        icp = i_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

			/* relaxation zrLV_RELAX5_P2_ON_D */
            *(ptr_V + icp) = 
                       (*(ptr_V + icp + size_col) * 4.0 
                        +
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        /
                        6.0;

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* loop P2 */


    /*=== P3 ===*/

    /* new ipc value */
    icp = size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

		/* relaxation zrLV_RELAX5_P3_ON_D */
        *(ptr_V + icp) = 
                    (*(ptr_V + icp + size_col) * 4.0
                     +
                     *(ptr_V + icp -1)
                     )
                     *
                     0.2;

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
		
    }/* P3, if Bnd == 0*/


    
    /*=== loop P4-P5-P6 ===*/

	/* direction: bottom -> up */   
    for(i_row = 2; i_row <= size_row - 1; i_row++)
    {

		/*=== P4 ===*/
		
        /* new ipc value */
        icp = (i_row - 1) * size_col;

        if(*(ptr_Bnd + icp) == 0)
        {

            alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1) / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp + 1) / 3.0
                        )
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P4, if Bnd == 0*/


        /*=== loop P5 ===*/

        /* new values for each row */
        alpha = (double)(2 * i_row - 3) /  (double)(8 * i_row - 8);
        beta  = (double)(2 * i_row - 1) /  (double)(8 * i_row - 8);
        
		/* direction: left -> right */   
        for(i_col = 2; i_col <= size_col - 1; i_col++)
        {
            /* new icp value */
            icp = (i_row - 1) * size_col + i_col - 1;

            if(*(ptr_Bnd + icp) == 0)
            {
  
                tmp = *(ptr_V + icp);

                *(ptr_V + icp)
                          =
                        (
                        *(ptr_V + icp - 1)
                        +
                        *(ptr_V + icp + 1)
                        )
                        *
                        0.25
                        +
                        (
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        )
                        +
                        (
                        beta
                        *
                        *(ptr_V + icp + size_col)
                        );       

                eps = *(ptr_V + icp) - tmp;

                if(eps < 0.0)
                {
                    /* absolute value */
                    eps = -eps;
                }
    
                if(eps > eps_max)
                {
                    /* new eps_max */
                    eps_max = eps;
                }

            }/* P5, if Bnd == 0*/

        }/* loop P5 */


        /*=== P6 ===*/

        /* new icp value */
        icp = i_row * size_col - 1;

        if(*(ptr_Bnd + icp) == 0)
        {
            alpha = (double)(2 * i_row - 3)  / (double)(6 * i_row - 6);
            beta  = (double)(2 * i_row - 1)  / (double)(6 * i_row - 6);

            tmp = *(ptr_V + icp);

            *(ptr_V + icp)
                        =
                        *(ptr_V + icp - 1)
                        /
                        3.0
                        +
                        alpha
                        *
                        *(ptr_V + icp - size_col)
                        +
                        beta
                        *
                        *(ptr_V + icp + size_col);

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* P6, if Bnd == 0*/

    }/* loop P4 - P5 - P6 */


    /*=== P7 ===*/

    icp = (size_row - 1) * size_col;

    if(*(ptr_Bnd + icp) == 0)
    {

        alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
        beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);

        tmp = *(ptr_V + icp);

        *(ptr_V + icp) =
                    alpha
                    *
                    (
                    *(ptr_V + icp - size_col)
                    )
                    +
                    beta
                    *
                    (
                    *(ptr_V + icp + 1)
					);

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P7, if Bnd == 0*/


    /*=== loop P8 ===*/

	/* direction: left -> right */   
    for(i_col = 2; i_col <= size_col - 1; i_col++)
    {

        /*new value of ipc*/
        alpha = (double)(2 * i_row - 3) / (double)(6 * i_row - 7);
        beta  = (double)(2 * i_row - 2) / (double)(6 * i_row - 7);

        icp = (size_row - 1) * size_col + i_col - 1;
 
        if(*(ptr_Bnd + icp) == 0)
        {
            tmp = *(ptr_V + icp);

            *(ptr_V + icp) = 
                        alpha
                        *
                        *(ptr_V + icp - size_col)  
                        +
                        beta
                        *         
                        (
                        *(ptr_V + icp - 1) 
                        +  
                        *(ptr_V + icp + 1)
                        );

            eps = *(ptr_V + icp) - tmp;

            if(eps < 0.0)
            {
                /* absolute value */
                eps = -eps;
            }
    
            if(eps > eps_max)
            {
                /* new eps_max */
                eps_max = eps;
            }

        }/* if Bnd == 0*/

    }/* loop P8 */


    /*=== P9 ===*/

    /* new ipc value */
    i_row = size_row;

    alpha = (double)(2 * i_row - 3) / (double)(4 * i_row - 5);
    beta  = (double)(2 * i_row - 2) / (double)(4 * i_row - 5);

    icp = size_row * size_col - 1;

    if(*(ptr_Bnd + icp) == 0)
    {
        tmp = *(ptr_V + icp);

        *(ptr_V + icp) = 
                    alpha
                    *
                    *(ptr_V + icp - size_col)
                    +
                    beta
                    *
                    *(ptr_V + icp - 1);

        eps = *(ptr_V + icp) - tmp;

        if(eps < 0.0)
        {
            /* absolute value */
            eps = -eps;
        }
    
        if(eps > eps_max)
        {
            /* new eps_max */
            eps_max = eps;
        }
    }/* P9, if Bnd == 0*/


    /* end */
    return eps_max;
	
}/* lbmn_relax_zr_D_iter_BR */


/*================================================================================*/

int
lbmn_relax_zr_D_set_point(
        double*        ptr_V,
        unsigned int*  ptr_Bnd,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col,
        unsigned int   i_row,/* node position (1,2, .. , n_row)*/
        unsigned int   i_col,/* 1,2,..,n_col*/
        double         voltage, /*input node values*/
        unsigned int   boundary)
{
	/* instructions */
	
	if(ptr_V == NULL ||
		ptr_Bnd == NULL ||
		size_row < 3 ||
		size_col < 3 
	)
	{
		return -1;
	}
	
    if(i_row > size_row)
    {
        return -2;
    }
    
    if(i_col > size_col)
    {
        return -3;
    }

	/* else input data seems to be correct */
	
    /*set voltage*/
    *(ptr_V +   (i_row - 1) * size_col + i_col - 1) = voltage;

    /* set boundary */
    *(ptr_Bnd + (i_row - 1) * size_col + i_col - 1) = boundary;

    return 0;
	
}/* lbmn_relax_zr_D_set_point */


/*=========================================================================*/


int
lbmn_relax_zr_D_printf_V_zr(
        double*        ptr_V,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col)
{
    /* declaration of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    double tmp = 0.0;

	/* instructions */
	
	if(ptr_V == NULL ||
		size_row < 3 ||
		size_col < 3
		)
	{
		return -1;
	}

    /* without axis*/
    for(i_row = size_row; i_row >1; i_row--)
    {
        for(i_col = 1; i_col <= size_col; i_col++)
        {
            tmp = *(ptr_V + (i_row-1) * size_col + i_col - 1);
            printf("%+18.15E ", tmp);
        }/*i_col*/
        
        printf("\n");

    }/*i_row*/

    /*axis + else */
    for(i_row = 1; i_row <= size_row; i_row++)
    {
        for(i_col = 1; i_col <= size_col; i_col++)
        {
            tmp = *(ptr_V + (i_row-1) * size_col + i_col - 1);
            printf("%+18.15E ", tmp);
        }/*i_col*/
        
        printf("\n");

    }/*i_row*/

	return 0;

}/* lbmn_relax_zr_D_printf_V_zr */


/*===============================================================*/


int
lbmn_relax_zr_D_printf_Bnd_zr(
        unsigned int*   ptr_Bnd,
        unsigned int    size_row,/* size of mesh */
        unsigned int    size_col)
{
    /* declartation of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    unsigned int tmp = 0.0;

	/* instructions */
	
	if(ptr_Bnd == NULL ||
		size_row < 3 ||
		size_col < 3
	)
	{
		return -1;
	}

    /* without axis*/
    for(i_row = size_row; i_row >1; i_row--)
    {
        for(i_col = 1; i_col <= size_col; i_col++)
        {
            tmp = *(ptr_Bnd + (i_row-1) * size_col + i_col - 1);
            printf("%d ", tmp);
        }/*i_col*/
        
        printf("\n");

    }/*i_row*/

    /*axis + else */
    for(i_row = 1; i_row <= size_row; i_row++)
    {
        for(i_col = 1; i_col <= size_col; i_col++)
        {
            tmp = *(ptr_Bnd + (i_row-1) * size_col + i_col - 1);
            printf("%d ", tmp);
        }/*i_col*/
        
        printf("\n");

    }/*i_row*/

	return 0;

}/* lbmn_relax_zr_D_printf_Bnd_zr */


/*=================================================================================*/


int
lbmn_relax_zr_D_find_V_min_max(
        double*         ptr_min,
        double*         ptr_max,
        double*         ptr_V,  
        unsigned int    size_row, /* size of mesh */
        unsigned int    size_col)
{
	/* declaration of variables */
	
    double tmp_min = 0.0;
    double tmp_max = 0.0;

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    double tmp = 0;

	/* instructions */
	
	if(ptr_min == NULL ||
		ptr_max == NULL ||
		ptr_V 	== NULL ||
		size_row < 3	||
		size_col < 3
	)
	{
		return -1;
	}
	
    /* find min / max */
    tmp_min = *(ptr_V + 0);
    tmp_max = *(ptr_V + 0);

    for(i_row = 1; i_row <= size_row; i_row++)
    {

        for(i_col = 1; i_col <= size_col; i_col++)
        {
            
            tmp = *(ptr_V + (i_row-1) * size_col + i_col - 1);

            if(tmp < tmp_min)
            {
                tmp_min = tmp;
            }

            if(tmp > tmp_max)
            {
                tmp_max = tmp;
            }

        }/*i_col*/

    }/* searching for min / max*/

    *ptr_min = tmp_min;
    *ptr_max = tmp_max;
	
    return 0;

}/*lbmn_relax_zr_D_find_V_min_max */


/*===========================================================================*/


int
lbmn_relax_zr_D_fprintf_V_netpbm_P3(
        const char* filename,
        double*        ptr_V,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col,
        int(*fptr_colormap)(/* pointer to colormap function */
                unsigned char* ptr_red,
                unsigned char* ptr_green,
                unsigned char* ptr_blue,
                double val,
                double val_min,
                double val_max))
{
    /* declaration of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;


    unsigned char tmp_red   = 0;
    unsigned char tmp_green = 0;
    unsigned char tmp_blue  = 0;

    double tmp_min = 0.0;
    double tmp_max = 0.0;

    FILE* ptr_file = NULL;

	/* instructions */
	
	if(filename == NULL ||
		ptr_V == NULL ||
		size_row < 3 ||
		size_col < 3 ||
		fptr_colormap == NULL
	)
	{
		return -1;
	}
	
    ptr_file = fopen (filename,"w");

    if (ptr_file == NULL)
    {
        /* could not open output file */
        return -2;
    }
    else
    {/*opened output file */

        /* write file header */

        fprintf (ptr_file, "P3\n %u %u\n 255\n", size_col, 2*size_row-1);
 
        if(ferror(ptr_file))
        {
            /* could not write header */
            fclose(ptr_file);
            return -3;
        }
        else
        {/* header written successfully */
            
            lbmn_relax_zr_D_find_V_min_max(
                            &tmp_min,
                            &tmp_max,
                            ptr_V,  
                            size_row,
                            size_col);

            /* LOOP 1 / without axis */
            for(i_row = size_row; i_row >1; i_row--)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                    if(fptr_colormap(    
                                    &tmp_red,
                                    &tmp_green,
                                    &tmp_blue,
                                    *(ptr_V + (i_row-1) * size_col + i_col - 1),
                                    tmp_min,
                                    tmp_max))
                    {
                        fclose(ptr_file);
                        return -4;
                    }
                    else
                    {/*colormap ok*/  
          
                        fprintf(ptr_file, "%3u %3u %3u\n",tmp_red, tmp_green, tmp_blue);

                        if(ferror(ptr_file))
                        {
                            fclose(ptr_file);
                            return -5;
                        }
                    }/*colormap ok*/

                }/*i_col*/

            }/*i_row*/


            /* LOOP 2 / axis + else */
            for(i_row = 1; i_row <= size_row; i_row++)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {


                    if(fptr_colormap(    
                                    &tmp_red,
                                    &tmp_green,
                                    &tmp_blue,
                                    *(ptr_V + (i_row-1) * size_col + i_col - 1),
                                    tmp_min,
                                    tmp_max))
                    {
                        fclose(ptr_file);
                        return -6;
                    }
                    else
                    {/* colormap ok */

                        fprintf(ptr_file, "%3u %3u %3u\n", tmp_red, tmp_green, tmp_blue);

                        if(ferror(ptr_file))
                        {
                            fclose(ptr_file);
                            return -7;

                        }/* return -7*/

                    }/* colormap ok */

                }/*i_col*/
                        
            }/*LOOP 2; i_row*/


            /* success */
            fclose(ptr_file);
                
        }/* header written successfully */
        
    }/* opened output file */

    return 0;
	
}/* lbmn_relax_zr_D_fprintf_V_netpbm_P3 */


/*==========================================================================*/


int
lbmn_relax_zr_D_fprintf_any_workbench_netpbm_P3(
        const char* filename,
        unsigned char* ptr_red,
		unsigned char* ptr_green,
		unsigned char* ptr_blue,
        unsigned int   size_row,/* size of mesh */
        unsigned int   size_col)
{
    /* declaration of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    FILE* ptr_file = NULL;

	/* instructions */

	if(filename == NULL ||
		ptr_red == NULL ||
		ptr_green == NULL ||
		ptr_blue == NULL ||
		size_row < 3 ||
		size_col < 3
	)
	{
		return -1;
	}
	
    ptr_file = fopen (filename,"w");

    if (ptr_file == NULL)
    {
        /* could not open output file */
        return -2;
    }
    else
    {/*opened output file */

        /* write file header */

        fprintf (ptr_file, "P3\n %u %u\n 255\n", size_col, 2*size_row-1);
 
        if(ferror(ptr_file))
        {
            /* could not write header */
            fclose(ptr_file);
            return -3;
        }
        else
        {/* header written successfully */
            

            /* LOOP 1 / without axis */
            for(i_row = size_row; i_row >1; i_row--)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                     
          
                        fprintf(ptr_file, "%3u %3u %3u\n",
								*(ptr_red 	+ (i_row-1)* size_col + i_col - 1),
								*(ptr_green + (i_row-1)* size_col + i_col - 1),
								*(ptr_blue 	+ (i_row-1)* size_col + i_col - 1));

                        if(ferror(ptr_file))
                        {
                            fclose(ptr_file);
                            return -4;
                        }

                }/*i_col*/

            }/*i_row*/


            /* LOOP 2 / axis + else */
            for(i_row = 1; i_row <= size_row; i_row++)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

					fprintf(ptr_file, "%3u %3u %3u\n",		
							*(ptr_red 	+ (i_row-1)* size_col + i_col - 1),
							*(ptr_green + (i_row-1)* size_col + i_col - 1),
							*(ptr_blue 	+ (i_row-1)* size_col + i_col - 1));
							
                        if(ferror(ptr_file))
                        {
                            fclose(ptr_file);
                            return -5;

                        }/* return -5*/

                }/*i_col*/
                        
            }/*LOOP 2; i_row*/


            /* success */
            fclose(ptr_file);
                
        }/* header written successfully */
        
    }/* opened output file */

    return 0;
	
}/* lbmn_relax_zr_D_fprintf_any_workbench_netpbm_P3 */


/*==================================================================*/


int
lbmn_relax_zr_D_fprintf_any_V_z(
        const char*     filename,
		double*         ptr_z,
        double*         ptr_V,
        unsigned int    size_col)
{
	/* declaration of variables */

	FILE * ptr_file;
    unsigned int  i = 0;

	/* instructions */
	
	if(filename == NULL ||
		ptr_z   == NULL ||
		ptr_V 	== NULL ||
        size_col < 3		
	)
	{
		return -1;
	}
	
    ptr_file = fopen (filename,"w");

    if(ptr_file == NULL)
    {
        return -2;
    }
    else
    {
        for (i = 1 ; i <= size_col; i++)
        {

			/* pair z <--> v(z)*/
            fprintf (ptr_file, "%+18.15E %+18.15E\n", *(ptr_z + i - 1), *(ptr_V + i - 1));

            if(ferror(ptr_file))
            {

                fclose(ptr_file);

                return -3;
            }
    

        }/* for */

        if(ferror(ptr_file))
        {
            fclose(ptr_file);
            return -4;
        }
    }

    return 0;
	
}/* lbmn_relax_zr_D_fprintf_any_V_z */


/*==================================================================*/


int
lbmn_relax_zr_D_fprintf_V_dat(
            const char* filename,
            double*        ptr_V,
            unsigned int   size_row,/* size of mesh */
            unsigned int   size_col)
{
    /* declaration of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    FILE* ptr_file = NULL;

	/* instructions */

    ptr_file = fopen (filename,"w");

	if(filename == NULL ||
		ptr_V   == NULL ||
		size_row < 3 ||
		size_col < 3
	)
	{
		return -1;
	}
	
    if (ptr_file == NULL)
    {
        /* could not open output file */
        return -2;
    }
    else
    {/*opened output file */

         

            /* LOOP 1 / without axis */
            for(i_row = size_row; i_row >1; i_row--)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                    fprintf(ptr_file, "%+18.15E ", *(ptr_V + (i_row-1) * size_col + i_col - 1));

                    if(ferror(ptr_file))
                    {
                        fclose(ptr_file);
                        return -3;
                    }
                }/*i_col*/


                fprintf(ptr_file, "\n");

                if(ferror(ptr_file))
                {
                    fclose(ptr_file);
                    return -4;
                }

            }/*i_row*/


            /* LOOP 2 / axis + else */
            for(i_row = 1; i_row <= size_row; i_row++)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                    fprintf(ptr_file, "%+18.15E ", *(ptr_V + (i_row-1) * size_col + i_col - 1));

                    if(ferror(ptr_file))
                    {
                        fclose(ptr_file);
                        return -5;
                    }


                }/*i_col*/
        
                fprintf(ptr_file, "\n");
              
                if(ferror(ptr_file))
                {
                    fclose(ptr_file);
                    return -6;
                }
            }/*LOOP 2; i_row*/


            /* success */
            fclose(ptr_file);

        
    }/* opened output file */

    return 0;
	
}/* lbmn_relax_zr_D_fprintf_V */


/*===================================================================*/

int
lbmn_relax_zr_D_fprintf_Bnd_dat(
            const char*     filename,
            unsigned int*   ptr_Bnd,
            unsigned int    size_row,/* size of mesh */
            unsigned int    size_col)
{
	/* declaration of variables */

    unsigned int i_row = 0;
    unsigned int i_col = 0;

    FILE* ptr_file = NULL;

	/* instructions */
	
	if(filename == NULL ||
	   ptr_Bnd  == NULL ||
	   size_row < 3     ||
	   size_col < 3
	)
	{
		return -1;
	}

    ptr_file = fopen (filename,"w");

    if (ptr_file == NULL)
    {
        /* could not open output file */
        return -1;
    }
    else
    {/*opened output file */

         

            /* LOOP 1 / without axis */
            for(i_row = size_row; i_row >1; i_row--)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                    fprintf(ptr_file, "%u ", *(ptr_Bnd + (i_row-1) * size_col + i_col - 1));

                    if(ferror(ptr_file))
                    {
                        fclose(ptr_file);
                        return -3;
                    }
                }/*i_col*/


                fprintf(ptr_file, "\n");

                if(ferror(ptr_file))
                {
                    fclose(ptr_file);
                    return -4;
                }

            }/*i_row*/

            /* LOOP 2 / axis + else */
            for(i_row = 1; i_row <= size_row; i_row++)
            {
                for(i_col = 1; i_col <= size_col; i_col++)
                {

                    fprintf(ptr_file, "%u ", *(ptr_Bnd + (i_row-1) * size_col + i_col - 1));

                    if(ferror(ptr_file))
                    {
                        fclose(ptr_file);
                        return -5;
                    }


                }/*i_col*/
        
                fprintf(ptr_file, "\n");
              
                if(ferror(ptr_file))
                {
                    fclose(ptr_file);
                    return -6;
                }
            }/*LOOP 2; i_row*/


            /* success */
            fclose(ptr_file);

        
    }/* opened output file */

    return 0;
	
}/* lbmn_relax_zr_D_fprintf_Bnd_dat */

/*====================================================================*/

int
lbmn_relax_zr_D_calculate_all_tab_V_z_h(
            double*     	ptr_V_z,
			double*			ptr_d1V_dz1_h,
			double*			ptr_d2V_dz2_h,
            double*     	ptr_V,
            unsigned int    size_col,
			double 			mesh_step_z)
{
	/* declaration of variables */
	
	unsigned int index = 0;
	
	/* instructions */
	
	printf("mesh step == %+f\n", mesh_step_z);
	
	/* check input data correctness */
	
	if(
		(ptr_V_z       == NULL) ||
		(ptr_d1V_dz1_h == NULL) ||
		(ptr_d2V_dz2_h == NULL) ||
		(ptr_V         == NULL) ||
		(size_col      < 3)
	)
	{/* inputa data is incorrect */
	
		return -1;
	
	}/*  inputa data is incorrect  */
	else
	{/*  inputa data is correct  */

		/********************************************************
		*														*
		*	do V_z (simple copy from tab_V mesh (1st row)     *
		*														*
		********************************************************/
		
		for(index = 1; index <= size_col; index ++)
		{
			*(ptr_V_z + index - 1) = *(ptr_V + index - 1);
		}


		/****************
		*				*
		* do d1V_dz1  *
		*				*
		****************/
		
		/*element 1 */
		index = 1;
		
		*(ptr_d1V_dz1_h + index - 1) = 0.0; 										

		
		/* element 2 .. size_col */
		
		for(index = 2; index <= size_col; index ++)
		{
			*(ptr_d1V_dz1_h + index - 1) =((*(ptr_V_z + index -1 )) - (*(ptr_V_z + index - 2 )))	/ mesh_step_z;
											
		}
		
		
		
		/*element size_col + 1 */
		index = size_col + 1;
		
		*(ptr_d1V_dz1_h + index - 1) = 0.0; 
		
		
		
		/****************
		*				*
		* do d2V_dz2  *
		*				*
		****************/
		

		/* element 1 .. size_col */
		
		for(index = 1; index <= size_col ; index ++)
		{
			*(ptr_d2V_dz2_h + index - 1) = ((*(ptr_d1V_dz1_h + index - 1 + 1))
										 -
										 (*(ptr_d1V_dz1_h + index  - 1))
										)
										/ mesh_step_z;
											
		}
		

		return 0;
		
	}/*  inputa data is correct  */
	
	
}/* lbmn_relax_zr_D_calculate_all_tab_V_z_h */


/*=============================================================================*/


int
lbmn_relax_zr_D_calculate_tab_V_z_2h(
			double*			ptr_d1V_dz1_2h,
			double*			ptr_d2V_dz2_2h,
            double*     	ptr_V,
            unsigned int    size_col,
			double 			mesh_step_z)
{
	/* declaration of variables */
	
	unsigned int index = 0;
	
	/* instructions */
	
	printf("mesh step == %+f\n", mesh_step_z);
	
	/* check input data correctness */
	
	if(
		(ptr_d1V_dz1_2h == NULL) ||
		(ptr_d2V_dz2_2h == NULL) ||
		(ptr_V          == NULL) ||
		(size_col       < 3)
	)
	{/* inputa data is incorrect */
	
		return -1;
	
	}/*  inputa data is incorrect  */
	else
	{/*  inputa data is correct  */


		/*****************
		*				 *
		* do d1V_dz1_2h  *
		*				 *
		******************/
		
		/*element 1 */
		index = 1;
		
		*(ptr_d1V_dz1_2h + index - 1) = 0.0; 										

		
		/* element 2 .. size_col - 1 */
		
		for(index = 2; index <= size_col - 1; index ++)
		{
			*(ptr_d1V_dz1_2h + index - 1) =((*(ptr_V + index - 1 + 1 )) - (*(ptr_V + index - 1 -1 )))	* 0.5 / mesh_step_z;
											
		}
		
		/*element size_col + 1 */
		index = size_col;
		
		*(ptr_d1V_dz1_2h + index - 1) = 0.0; 
		
		
		
		/*****************
		*				 *
		* do d2V_dz2_2h  *
		*				 *
		******************/
		

		/*element 1 */
		index = 1;
		
		*(ptr_d2V_dz2_2h + index - 1) = 0.0; 										

		
		/* element 2 .. size_col - 1 */
		
		for(index = 2; index <= size_col - 1; index ++)
		{
			*(ptr_d2V_dz2_2h + index - 1) =((*(ptr_d1V_dz1_2h + index - 1 + 1 )) - (*(ptr_d1V_dz1_2h + index - 1 -1 )))	* 0.5 / mesh_step_z;
											
		}
		
		/*element size_col + 1 */
		index = size_col;
		
		*(ptr_d2V_dz2_2h + index - 1) = 0.0; 
		

		

		return 0;
	}/*  inputa data is correct  */
	
	
}/* lbmn_relax_zr_D_calculate_all_tab_V_z_2h */


/*=======================================================================*/


int
lbmn_relax_zr_D_create_image_workbench_bw(
            unsigned char* 	ptr_red,
			unsigned char*	ptr_green,
			unsigned char*	ptr_blue,
            unsigned int   	size_row,/* size of mesh */
            unsigned int   	size_col,
			unsigned int*	ptr_Bnd,/* mesh of boundaries*/
			unsigned char	axis_colour_red,
			unsigned char	axis_colour_green,
			unsigned char	axis_colour_blue)
{
	/* declaration of variables */

	unsigned int i_row;
	unsigned int i_col;

	unsigned int point_Bnd = 0;

	/* instructions */
	
	if(
		ptr_red 	== NULL ||
		ptr_green	== NULL ||
		ptr_blue	== NULL ||
		ptr_Bnd 	== NULL ||
		size_row < 3   		||
		size_col < 3		
	  )
	{
		return -1;
	}
	

	/* loop 1 - on optical axis */
	
	for(i_col = 1; i_col <= size_col; i_col++)
	{
		*(ptr_red 	+ i_col - 1) = axis_colour_red;
		*(ptr_green + i_col - 1) = axis_colour_green;
		*(ptr_blue 	+ i_col - 1) = axis_colour_blue;
	}

	
	/* loop 2 - over optical axis 	*/
	
	for(i_row = 2; i_row <= size_row; i_row++)
	{
		for(i_col = 1; i_col <= size_col; i_col++)
		{
			/* determine electrode point */
			point_Bnd =  *(ptr_Bnd + (i_row-1)* size_col + i_col - 1);
			
			
			if(point_Bnd != 0)
			{/*electrode point*/
		
				/* set colour to black */
				*(ptr_red 	+ (i_row-1)* size_col + i_col - 1) = 0;
				*(ptr_green + (i_row-1)* size_col + i_col - 1) = 0;
				*(ptr_blue	+ (i_row-1)* size_col + i_col - 1) = 0;
				
				
			}/* electrode point*/
			
			else
			{/*vacuum*/
		
				/* set colour to white */				
				*(ptr_red 	+ (i_row-1)* size_col + i_col - 1) = 255;
				*(ptr_green + (i_row-1)* size_col + i_col - 1) = 255;
				*(ptr_blue	+ (i_row-1)* size_col + i_col - 1) = 255;

			}/*vacuum*/
		}
	}

	return 0;
	
}/* lbmn_relax_zr_D_create_image_workbench_bw */

/*==============================================================================*/

int
lbmn_relax_zr_D_create_image_workbench_colours(
            unsigned char* 	ptr_red,
			unsigned char*	ptr_green,
			unsigned char*	ptr_blue,
            unsigned int   	size_row,/* size of mesh */
            unsigned int   	size_col,
			unsigned int*	ptr_Bnd,/* mesh of boundaries*/
			unsigned char*	ptr_electrode_colours_red,
			unsigned char*	ptr_electrode_colours_green,
			unsigned char*	ptr_electrode_colours_blue,
			unsigned int	size_electrode_colours,
			unsigned char	axis_colour_red,
			unsigned char	axis_colour_green,
			unsigned char  	axis_colour_blue)
{
	/* declaration of variables */
	
	unsigned int i_row;
	unsigned int i_col;
	
	unsigned int point_Bnd = 0;

	/* instructions */
	
	if(
		ptr_red 	== NULL ||
		ptr_green	== NULL ||
		ptr_blue	== NULL ||
		ptr_Bnd 	== NULL ||
		size_row < 3   		||
		size_col < 3		||
		ptr_electrode_colours_red 	== NULL ||
		ptr_electrode_colours_green == NULL ||
		ptr_electrode_colours_blue 	== NULL ||
		size_electrode_colours == 0
		
	  )
	{
		return -1;
	}
	
	
	/*loop 1 - on optical axis*/
	
	for(i_col = 1; i_col <= size_col; i_col++)
	{
		*(ptr_red 	+ i_col - 1) = axis_colour_red;
		*(ptr_green + i_col - 1) = axis_colour_green;
		*(ptr_blue 	+ i_col - 1) = axis_colour_blue;
	}

	
	/* loop 2 - over optical axis 	*/
	
	for(i_row = 2; i_row <= size_row; i_row++)
	{
		for(i_col = 1; i_col <= size_col; i_col++)
		{
			/* determine electrode point */
			point_Bnd =  *(ptr_Bnd + (i_row-1)* size_col + i_col - 1);
			
			
			/* set colour */

			*(ptr_red 	+ (i_row-1)* size_col + i_col - 1) = *(ptr_electrode_colours_red   + point_Bnd);
			*(ptr_green + (i_row-1)* size_col + i_col - 1) = *(ptr_electrode_colours_green + point_Bnd);
			*(ptr_blue	+ (i_row-1)* size_col + i_col - 1) = *(ptr_electrode_colours_blue  + point_Bnd);
		
		}
	}

	return 0;
	
}/* lbmn_relax_zr_D_create_image_workbench_colours */

/* ==== end of file lbmn_relax_zr_D.c ==== */